#!/data/salomonis2/LabFiles/Frank-Li/refactor/neo_env/bin/python3.7

import numpy as np
import pandas as pd
import os
import sys
import pickle
import matplotlib.pyplot as plt
import seaborn as sns
from lifelines import KaplanMeierFitter
from lifelines.statistics import logrank_test
import statsmodels.stats.multitest as ssm
from scipy.stats import mannwhitneyu,pearsonr,spearmanr
from tqdm import tqdm
from copy import deepcopy
import multiprocessing as mp
import subprocess

'''
this script contains survival analysis, mutation analysis
'''

import matplotlib as mpl
mpl.rcParams['pdf.fonttype'] = 42
mpl.rcParams['ps.fonttype'] = 42
mpl.rcParams['font.family'] = 'Arial'

def plot_umap_neoantigen(df_path,outdir):
    df = pd.read_csv(df_path,sep='\t',index_col=0)
    fig,ax = plt.subplots()
    ax.scatter(df['umap_x'],df['umap_y'],color=['red' if item == 'high' else 'blue' for item in df['identity']],s=3)
    from matplotlib.lines import Line2D
    ax.legend(handles=[Line2D([],[],marker='o',linestyle='',color=i) for i in ['red','blue']],labels=['Shared Neoantigen','Unique Neoantigen'],bbox_to_anchor=(1,1),loc='upper left',frameon=False)
    ax.set_xlabel('umap_x')
    ax.set_ylabel('umap_y')
    plt.savefig(os.path.join(outdir,'mer_umap.pdf'),bbox_inches='tight')
    plt.close()

def survival_regression(freq,remove_quote,rename_func,survival,pea,outdir='.',cores=None,mode='binary',survival_duration='OS.time',survival_event='OS',n_effective_obs=3):
    '''
    conduct cox regression to identify neoantigens whose parental junctions expression is significantly associated with survival

    :param freq: string, path to the T antigen generated frequency dataframe
    :param remove_quote: bool, whether to remove the quote of samples column for freq df or not
    :param rename_func: function, a function to rename columns in freq and event annotation file to be consistent with survival file
    :param survival: string, the path to the survival dataframe
    :param pea: string, the path to the altanalyze generated event annotation file
    :param outdir: string, the output directory
    :param cores: None or int, how many cores to use for computation
    :param mode: binary or psi, if binary, we test whether having a neoantigen can predict survival, psi means whether its parental junction's psi value is associated with survival
    :param survival_duration: string, the column name for duration in survival dataframe
    :param survival_event: string, the column name for event in survival dataframe
    :param n_effective_obs: int, default is 3, at least three obs whose event=1, otherwise all obs are right-sensored

    Examples::

        snaf.downstream.survival_regression(freq='result/frequency_stage3_verbosity1_uid_gene_symbol_coord_mean_mle.txt',remove_quote=True,
                                            rename_func=lambda x:'-'.join(x.split('-')[:4]),survival='TCGA-SKCM.survival.tsv',
                                            pea='Hs_RNASeq_top_alt_junctions-PSI_EventAnnotation.txt',outdir='result/survival')
    '''
    # reformat psi matrix for further extraction
    ea = pd.read_csv(pea,sep='\t',index_col='UID').iloc[:,10:]
    ea.index = [':'.join(item.split('|')[0].split(':')[1:]) for item in ea.index]
    ea = ea.loc[np.logical_not(ea.index.duplicated()).tolist(),:]
    ea.rename(columns=rename_func,inplace=True)
    # format freq matrix
    freq = pd.read_csv(freq,index_col=0,sep='\t')
    from ast import literal_eval
    if remove_quote:
        freq['samples'] = [literal_eval(item) for item in freq['samples']]
    # read in survival data
    survival = pd.read_csv(survival,index_col=0,sep='\t')
    survival = survival.loc[:,[survival_duration,survival_event]]
    # start to calculate association for each neoantigen 
    # before spawn, let's rename all the samples
    freq['samples'] = [[rename_func(s) for s in ss] for ss in freq['samples']]
    if cores is None:
        cores = mp.cpu_count()
    pool = mp.Pool(processes=cores)
    print('{} subprocesses have been spawned'.format(cores))
    sub_dfs = split_df_to_chunks(freq,cores=cores)
    if mode == 'binary':
        r = [pool.apply_async(func=survival_regression_binary_atomic,args=(sub_df,ea,survival,survival_duration,survival_event,n_effective_obs,)) for sub_df in sub_dfs]  
        pool.close()
        pool.join()
        df_data = []
        for collect in r:
            result = collect.get()
            df_data.extend(result)
    elif mode == 'psi':
        r = [pool.apply_async(func=survival_regression_psi_atomic,args=(sub_df,ea,survival,survival_duration,survival_event,n_effective_obs,)) for sub_df in sub_dfs]  
        pool.close()
        pool.join()
        df_data = []
        for collect in r:
            result = collect.get()
            df_data.extend(result)    
    final_df = pd.DataFrame.from_records(data=df_data,columns=['uid','pep','junc','n_valid_sample','wald_z_score','wald_p_value'])
    final_df.set_index(keys='uid',inplace=True)
    final_df.to_csv(os.path.join(outdir,'survival_regression_{}_final_results.txt'.format(mode)),sep='\t')

def split_df_to_chunks(df,cores=None):
    df_index = np.arange(df.shape[0])
    if cores is None:
        cores = mp.cpu_count()
    sub_indices = np.array_split(df_index,cores)
    sub_dfs = [df.iloc[sub_index,:] for sub_index in sub_indices]
    return sub_dfs

def survival_regression_binary_atomic(freq,ea,survival,survival_duration,survival_event,n_effective_obs):
    df_data = []
    from lifelines import CoxPHFitter
    for uid,ss in tqdm(zip(freq.index,freq['samples']),total=freq.shape[0]):
        pep,junc = uid.split(',')
        tmp = ea.loc[junc,:].to_frame()
        tmp.columns = ['psi']
        # binary, as psi is variable, we just want to know if they are present or not
        tmp.loc[ss,['psi']] = 1
        not_ss = list(set(tmp.index).difference(set(ss)))
        tmp.loc[not_ss,['psi']] = 0
        operation_df = tmp.join(other=survival,how='inner')
        operation_df = operation_df.dropna()  # actually unneccessary, because if a neojunction can give rise to neoantigen, so neojunction psi should be above an appreciable level
        n_valid_sample = len(ss)
        try:
            assert operation_df.loc[operation_df[survival_event]==1,:].shape[0] >= n_effective_obs
        except AssertionError:
            wald_z_score, wald_p_value = 'n_effective_obs<3','n_effective_obs<3'
            df_data.append((uid,pep,junc,n_valid_sample,wald_z_score,wald_p_value))
        else:
            cph = CoxPHFitter()
            try:
                cph.fit(operation_df,duration_col=survival_duration,event_col=survival_event)
                summary = cph.summary
                wald_z_score, wald_p_value = summary.loc['psi',['z','p']].tolist()
            except:
                wald_z_score, wald_p_value = 'convergence_error','convergence_error'
            df_data.append((uid,pep,junc,n_valid_sample,wald_z_score,wald_p_value))
    return df_data    

def survival_regression_psi_atomic(freq,ea,survival,survival_duration,survival_event,n_effective_obs):
    df_data = []
    from lifelines import CoxPHFitter
    for uid,ss in tqdm(zip(freq.index,freq['samples']),total=freq.shape[0]):
        pep,junc = uid.split(',')
        tmp = ea.loc[junc,ss].to_frame()
        tmp.columns = ['psi']
        operation_df = tmp.join(other=survival,how='inner')
        if uid == 'RETDFKMKF,ENSG00000167291:E38.6-E39.1':
            operation_df.to_csv('tmp_test_old.txt',sep='\t');sys.exit('stop')
        operation_df = operation_df.dropna()  # actually unneccessary, because if a neojunction can give rise to neoantigen, so neojunction psi should be above an appreciable level
        n_valid_sample = operation_df.shape[0]
        try:
            assert operation_df.loc[operation_df[survival_event]==1,:].shape[0] >= n_effective_obs
        except AssertionError:
            wald_z_score, wald_p_value = 'n_effective_obs<3','n_effective_obs<3'
            df_data.append((uid,pep,junc,n_valid_sample,wald_z_score,wald_p_value))
        else:
            cph = CoxPHFitter()
            try:
                cph.fit(operation_df,duration_col=survival_duration,event_col=survival_event)
                summary = cph.summary
                wald_z_score, wald_p_value = summary.loc['psi',['z','p']].tolist()
            except:
                wald_z_score, wald_p_value = 'convergence_error','convergence_error'
            df_data.append((uid,pep,junc,n_valid_sample,wald_z_score,wald_p_value))
    return df_data

def plot_event_type(pea,uids,rel,outdir):
    '''
    plot the event type based on provided uid list

    :param pea: string, the path to the EventAnnotation file
    :param uids: dict, key is the name of each category, value is a list of non-overlapping uids
    :param rel: bool, whether to compute relative or absolute value
    :param outdir: string, the output directory

    Examples::

        snaf.downstream.plot_event_type(pea='Hs_RNASeq_top_alt_junctions-PSI_EventAnnotation.txt',uids={'common':common_uid,'unique':unique_uid},rel=True,outdir='result_new')
    '''
    # {nan, "alt5'-alt3'", "alt-3'|cassette-exon", 'intron-retention', "alt-3'", "alt-5'", "alt-5'|cassette-exon", 'trans-splicing', 'altPromoter', 'alt-C-term', 'cassette-exon'}
    valid_type = ["cassette-exon","alt-3'","alt-5'","intron-retention","alt-C-term","altPromoter","trans-splicing"]
    valid_color = {"cassette-exon":'#3F61A6',
                    "alt-3'":'#07B2D9',
                    "alt-5'":'#F27B35',
                    "intron-retention":'#64BF4B',
                    "alt-C-term":'#A5A4A4',
                    "altPromoter":'#F2A413',
                    "trans-splicing":'#9F69AC'}
    df = pd.read_csv(pea,sep='\t')
    df['foreground'] = [':'.join(item.split('|')[0].split(':')[1:]) for item in df['UID']]
    evc_list = []
    for k,uid in uids.items():
        evc = df.loc[df['foreground'].isin(set(uid)),:]['EventAnnotation'].value_counts()
        evc.name = k
        evc_list.append(evc)
    evc_all = pd.concat(evc_list,axis=1,join='outer')
    cond = []
    for item in evc_all.index:
        if item in set(valid_type):
            cond.append(True)
        else:
            cond.append(False)
    evc_all = evc_all.loc[cond,:]
    if rel:
        evc_all = evc_all / evc_all.sum(axis=0)
    # plot
    fig,ax = plt.subplots()
    current_left = np.array([0] * evc_all.shape[1])
    for i in range(evc_all.shape[0]):
        v = evc_all.iloc[i,:].values
        t = evc_all.iloc[i,:].name
        ax.barh(y=np.arange(evc_all.shape[1]),width=v,height=0.5,left=current_left,color=valid_color[t])
        current_left = current_left + v
    import matplotlib.patches as mpatches
    ax.legend(handles=[mpatches.Patch(color=i) for i in valid_color.values()],labels=list(valid_color.keys()),frameon=False,loc='upper left',bbox_to_anchor=(0,-0.2),ncol=2)
    ax.set_yticks(np.arange(evc_all.shape[1]))
    ax.set_yticklabels(evc_all.columns.tolist(),fontsize=4)
    ax.grid(alpha=0.2)
    plt.savefig(os.path.join(outdir,'event_type_plot.pdf'),bbox_inches='tight')
    plt.close()
    return evc_all

def get_coverage(t_result,allele,outdir='.'):
    '''
    Get the population coverage of certain neoantigen, based on the different HLA allele frequency from
    US bone marrow registry. The 21 race code can be found here (https://www.sciencedirect.com/science/article/pii/S0198885913001821?via%3Dihub#t0005)

    :param t_result: string, the path to the T_antigen_candidates_all.txt file
    :param allele: string, either A, B or C
    :param outdir: string, default is current directory

    Examples::

        snaf.downstream.get_coverage(t_result='result_new/T_candidates/T_antigen_candidates_all.txt',allele='A')


    .. csv-table:: annotation txt file
        :file: ../docs/_static/coverage.csv
        :widths: 10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10
        :header-rows: 1   

    '''


    # race code table: https://www.sciencedirect.com/science/article/pii/S0198885913001821?via%3Dihub#t0005
    # frequency US registry: https://pubmed.ncbi.nlm.nih.gov/23806270/
    # download supp table5 as the dictionary
    import re
    df = pd.read_csv(os.path.join(os.path.dirname(os.path.abspath(__file__)),'HLA_Allele_frequency_21_populations.csv'),header=0,skiprows=[1,2]).iloc[:,1:].set_index(keys='Unnamed: 1')
    df.index.name = 'allele'
    df.dropna(axis=0,inplace=True)
    pat = re.compile(r'[ABC]\*\d+:\d+')
    col = []
    for item in df.index:
        match = re.search(pat,item)
        if match:
            col.append(match.group(0))
        else:
            col.append(item)
    df.index = col
    dic = df.to_dict('index')  # {A*01:01:{race:freq,race:freq}}

    # focus on each allele
    dic = {k:v for k,v in dic.items() if k[0] == allele}

    t_result = pd.read_csv(t_result,sep='\t')
    with open(os.path.join(outdir,'T_coverage_analysis_{}.txt'.format(allele)),'w') as f:
        f.write('peptide\thlas\tuid\tsymbol\tcoord\ttumor_specificity_mean\ttumor_specificity_mle\t{}\n'.format('\t'.join(df.columns.tolist())))
        for pep,sub_df in t_result.groupby(by='peptide'):
            tmp = sub_df.iloc[0]
            uid,symbol,coord,tsmean,tsmle = tmp['uid'],tmp['symbol'],tmp['coord'],tmp['tumor_specificity_mean'],tmp['tumor_specificity_mle']
            unique_hlas = [item.split('-')[1] for item in list(set(sub_df['hla'].tolist()))]
            accum = np.empty(shape=(df.shape[1],len(unique_hlas)),dtype=np.float64)
            for i,hla in enumerate(unique_hlas):
                try:
                    tmp = list(dic[hla].values())
                except KeyError:   # allele not present in the database
                    tmp = np.zeros((df.shape[1],))
                accum[:,i] = tmp
            accum = [str(item) for item in accum.sum(axis=1).tolist()]
            f.write('{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\n'.format(pep,','.join(unique_hlas),uid,symbol,coord,tsmean,tsmle,'\t'.join(accum)))
    


def prepare_DEG_analysis(burden_path,patient_strat_path,rename_func,outdir='.',encoding={'low':'1','high':'2'}):
    '''
    This function generate groups and comps file for performing AltAnalyze Differential Gene Analysis using empirical Bayes

    :param burden_path: string, the path to the burden file generated by T pipeline
    :param patient_strat_path: string, the path to the patient stratification file returned by survival_analysis function
    :param rename_func: function, it will be used to transform sample_id in burden_path (same as exp file) to sample_id in stat file (as it is adjusted for survival file)
    :param outdir: string, the folder for generating comps and groups files
    :param encoding: dict, the encoding to specificy how to transform identity string and what identities will be used for the DEG analysis


    Examples::

        snaf.downstream.prepare_DEG_analysis('result/burden_stage3.txt','result/survival/burden3_patient_high_low_group.txt',
                                             '/data/salomonis2/NCI-R01/TCGA-SKCM/TCGA_SKCM_hg38/ExpressionInput/exp.TCGA-SKCM.txt',
                                             outdir='result/survival',encoding={'low':'1','high':'2'})


    '''
    burden3 = pd.read_csv(burden_path,sep='\t',index_col=0)
    old_to_new = {}
    new_to_old = {}
    for x in burden3.columns:
        new_x = rename_func(x)
        old_to_new[x] = new_x
        new_to_old[new_x] = x
    strat = pd.read_csv(patient_strat_path,sep='\t',index_col=0)
    ident_dict = encoding
    with open(os.path.join(outdir,'comps.txt'),'w') as f:
        f.write('{}\t{}\n'.format(list(ident_dict.values())[0],list(ident_dict.values())[1]))
    with open(os.path.join(outdir,'groups.txt'),'w') as f:
        for row in strat.itertuples():
            if row.identity == list(ident_dict.keys())[0] or row.identity == list(ident_dict.keys())[1]:
                string = new_to_old[row.Index] + '\t' + ident_dict[row.identity] + '\t' + row.identity
                f.write('{}\n'.format(string))

    
def visualize_DEG_result(result_path,mode,outdir='.',genes_to_highlight=[],up_cutoff=0.8,down_cutoff=-0.8,xlims=(-2,2),ylims=None):
    '''
    Generate volcano plot for the AltAnalyze DEG results

    :param result_path: string, the path to the AltAnalyze DEG results
    :param mode: string, either interactive or static
    :param outdir: string, the folder where the plots will go into
    :param genes_to_highlight: list, the gene names to highlight in the plot, only for static mode
    :param up_cutoff: float, the up-regulated cutoff for logfoldchange, only for static mode
    :param down_cutoff: float, the down-regulated cutoff for logfoldchagne, only for static mode
    :param xlims: None or tuple, force the xlims of the volcano plot, only for static mode
    :param ylims: None or tuple, force the ylims of the volcano plot, only for static mode

    Examples::

        snaf.downstream.visualize_DEG_result('/data/salomonis2/LabFiles/Frank-Li/neoantigen/revision/TCGA_melanoma/result/survival/DEGs-LogFold_0.58_adjp/GE.low_vs_high.txt',
                                             mode='static',outdir='.',genes_to_highlight=['LST1','HCST','IL32','CD3D','S100A8','MZB1','IGLC4','ADAM10','ARFGEF2','MIB1','KIF3B','TNPO1','PTPN11','ANKRD52','TGFBR1'])

    '''
    
    df = pd.read_csv(result_path,sep='\t',index_col=0).loc[:,['LogFold','adjp','Symbol']]
    df['minus_log10_adjp'] = np.negative(np.log10(df['adjp'].values))
    df.drop_duplicates(subset=['Symbol'],inplace=True)
    problematics = list(set(genes_to_highlight).difference(set(df['Symbol'].values)))
    print('{} are not in DEG results'.format(problematics))
    if mode == 'interactive':
        import plotly.graph_objects as go
        node_x = []
        node_y = []
        node_text = []
        for row in df.itertuples():
            node_x.append(float(row.LogFold))
            node_y.append(float(row.minus_log10_adjp))
            node_text.append(row.Symbol)
        node_trace = go.Scatter(name='nodes',x=node_x,y=node_y,mode='markers',hoverinfo='text',text=node_text,marker={'color':'blue','size':5})
        fig_layout = go.Layout(showlegend=False,title='DEG',xaxis=dict(title_text='logfold'),yaxis=dict(title_text='-log10_adjp'))
        fig = go.Figure(data=[node_trace],layout=fig_layout)
        fig.write_html(os.path.join(outdir,'volcano_plot_DEG_interactive.html'),include_plotlyjs='cdn')
    elif mode == 'static':
        from adjustText import adjust_text
        from matplotlib.patches import ArrowStyle
        node_x = []
        node_y = []
        node_text = []
        node_color = []
        for row in df.itertuples():
            node_x.append(float(row.LogFold))
            node_y.append(float(row.minus_log10_adjp))
            node_text.append(row.Symbol)
            if float(row.LogFold) > up_cutoff:
                node_color.append('#F22727')
            elif float(row.LogFold) < down_cutoff:
                node_color.append('#41D2F2')
            else:
                node_color.append('#CFCFD0')
        fig,ax = plt.subplots()
        ax.scatter(node_x,node_y,s=2,c=node_color)  
        ax.spines['left'].set_position('center')
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.set_xlabel('logfold')
        ax.set_ylabel('-log10(adjp)')
        if xlims is not None:
            ax.set_xlim(xlims)
        if ylims is not None:
            ax.set_ylim(ylims)
        geneset = set(genes_to_highlight)
        texts = [ax.text(node_x[i],node_y[i],node_text[i],ha='center',va='center',fontsize=5) for i in range(len(node_text)) if node_text[i] in geneset]
        adjust_text(texts,arrowprops=dict(arrowstyle='-',color='#CFCFD0'))
        plt.savefig(os.path.join(outdir,'volcano_plot_DEG_static.pdf'),bbox_inches='tight');plt.close()


def prepare_GO_analysis(result_path,lc_cutoff=0.5,direction='>',adjp_cutoff=0.05,n=None,sortby='adjp',ascending=True,outdir='.'):
    '''
    Generate gene list for GO analysis

    :param result_path: string, the path to the AltAnalyze generated DEG result
    :param lc_cutoff: float, the cutoff above which will be considered desirable genes
    :param direction: string, either ">" or "<", indicating whether to use greater than or less than of lc_cutoff
    :param adjp_cutoff: float, the cutoff below which will be considered desirable genes
    :param n: None or int, the number of genes to contrain for GO analysis
    :param sortby: string, 'adjp' or 'LogFold'
    :param ascending: bool, sort in ascending order or not
    :param outdir: string, the directory to write outputs

    Examples::

        snaf.downstream.prepare_GO_analysis('result/survival/DEGs-LogFold_0.58_adjp/GE.low_vs_high.txt',outdir='result/survival')


    '''
    df = pd.read_csv(result_path,sep='\t',index_col=0).loc[:,['LogFold','adjp','Symbol']]
    df.drop_duplicates(subset=['Symbol'],inplace=True)
    if direction == '>':
        valid_df = df.loc[(df['LogFold']>lc_cutoff) & (df['adjp']<adjp_cutoff),:]
    elif direction == '<':
        valid_df = df.loc[(df['LogFold']<lc_cutoff) & (df['adjp']<adjp_cutoff),:]
    print('Total genes are {}, set n as {}, sort by {}, ascending {}'.format(valid_df.shape[0],n,sortby, ascending))
    valid_df = valid_df.sort_values(by=sortby,axis=0,ascending=ascending)
    valid_df.to_csv(os.path.join(outdir,'gene_dataframe.txt'),sep='\t')
    gene_list = valid_df['Symbol'].tolist()
    if n is None:
        n = len(gene_list)
    else:
        if n > len(gene_list):
            n = len(gene_list)
    with open(os.path.join(outdir,'gene_list.txt'),'w') as f:
        c = 0
        for i,gene in enumerate(gene_list):
            if c < n:
                f.write('{}\n'.format(gene))
                c += 1

def visualize_GO_result(path_list,skiprows_list,category_list,mode='interactive',outdir='',ontology_to_highlight={},
                        xlims=None,ylims=None):
    '''
    Visualize the GO analysis rssults

    :param path_list: list, each element points to the GO-Elite result txt file under ORA folder
    :param skiprows_list: list, each element represents the number of lines needing to be skipped for each result file
    :param category_list: list, each element points to the column name in each result file that represent GO terms
    :param mode: string, interactive or static
    :param outdir: string, the output directory
    :param ontology_to_highlight: dict, key is the GO terms in the result, values are the corresponding text you want them to be displayed on the plot
    :param xlims: None or tuple.
    :param ylims: None or tuple.

    Examples::

        snaf.downstream.visualize_GO_result(path_list=['result/survival/GO_Elite_result_GeneOntology/GO-Elite_results/CompleteResults/ORA/archived-20221101-142116/gene_list-GO.txt','result/survival/GO_Elite_result_BioMarkers/GO-Elite_results/CompleteResults/ORA/archived-20221101-142037/gene_list-BioMarkers.txt'],
                                    skiprows_list=[17,16],category_list=['Ontology Name','Gene-Set Name'],outdir='result/survival',
                                    mode='static',ontology_to_highlight={'Adult Peripheral Blood Activated T cell (PMID32214235 top 100)':'T cells','antigen binding':'antigen binding','complement activation':'Complement Activation','immune response':'immune response','humoral immune response':'humoral immune response'},ylims=(10e-50,10e-1))

    '''
    dfs = []
    for path,skiprows,column in zip(path_list,skiprows_list,category_list):
        df = pd.read_csv(path,sep='\t',index_col=column,skiprows=skiprows).loc[:,['Z Score','AdjustedP']]
        df.rename(columns={'Z Score':'z_score'},inplace=True)
        dfs.append(df)
    df_all = pd.concat(dfs,axis=0)
    if mode == 'interactive':
        import plotly.graph_objects as go
        node_x = []
        node_y = []
        node_text = []
        for row in df_all.itertuples():
            node_x.append(float(row.z_score))
            node_y.append(float(row.AdjustedP))
            node_text.append(row.Index)
        node_trace = go.Scatter(name='nodes',x=node_x,y=node_y,mode='markers',hoverinfo='text',text=node_text,marker={'color':'blue','size':5})
        fig_layout = go.Layout(showlegend=False,title='DEG',xaxis=dict(title_text='z_score',range=xlims),yaxis=dict(title_text='FDR_p',range=ylims))
        fig = go.Figure(data=[node_trace],layout=fig_layout)
        fig.write_html(os.path.join(outdir,'GO_plot_interactive.html'),include_plotlyjs='cdn')
    elif mode == 'static':
        from adjustText import adjust_text
        from matplotlib.patches import ArrowStyle
        node_x = []
        node_y = []
        node_text = []
        for row in df_all.itertuples():
            node_x.append(float(row.z_score))
            node_y.append(float(row.AdjustedP))
            node_text.append(row.Index)
        fig,ax = plt.subplots()
        ax.scatter(node_x,node_y,s=2,c='#3D52A3')  
        ax.spines['left'].set_position(('axes',0.3))
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.set_xlabel('GO-Elite Z-Score')
        ax.set_ylabel('FDR_p')
        ax.set_yscale('log')
        if xlims is not None:
            ax.set_xlim(xlims)
        if ylims is not None:
            ax.set_ylim(ylims)
        pathway_set = set(ontology_to_highlight.keys())
        texts = [ax.text(node_x[i],node_y[i],ontology_to_highlight[node_text[i]],ha='center',va='center',fontsize=3) for i in range(len(node_text)) if node_text[i] in pathway_set]
        adjust_text(texts,arrowprops=dict(arrowstyle='-',color='#CFCFD0'))
        plt.savefig(os.path.join(outdir,'GO_plot_static.pdf'),bbox_inches='tight');plt.close()

            

        
def prepare_sashimi_plot(bam_path_list,bai_path_list,outdir,sif_anno_path, bam_contig_rename, query_region, skip_copy=False, min_junction=3, width=10, ann_height=4, height=2, fix_y_scale=True, task_name=''):
    '''
    Given a bunch of bam file and cognate bai files, we want to generate sashimi plot in an automatic way using ggsashimi package singularity image, this image is pulled using::

        singularity build ggsashimi.sif docker://gencode.v36.annotation.gtf
    
    Now specifiy a folder (outdir), we are going to copy bam and bai to a subdirectory called ggsashimi_bams, automatically generate input_bams.tsv and palette.txt file,
    and then return a string to run singularity images::

        singularity run -W $PWD -B $PWD:$PWD ggsashimi.sif -b input_bam.tsv -c chr10:27040584-27048100 -C 3 -M 5 --width 10 --fix-y-scale -g gencode.v36.annotation.gtf --ann-height=4 --height=2 -P palette.txt

    :param bam_path_list: list, each element represents the path to the bam file
    :param bai_path_list: list, each element represents the path to the bai file
    :param outdir: string, the output directory where we are going to execute singularity run and get the plot, we prefer non-existing path or empty existing folder
    :param sif_anno_path: string, the folder where ggsashimi.sif and gencode.v36.annotation.gtf reside
    :param bam_contig_rename: list of boolean, whether to convert the respective bam contig from 10 to chr10, require the samtools are loaded
    :param query_region: string, format like chr10:27040584-27048100
    :param skip_copy: boolean, whether to skip copying step
    :param min_junction: int, the minimum number of junction to show sashimi line
    :param width: float, the width in pct for the plot
    :param ann_height: float, the height in pct for the annotation
    :param height: float, the height in pct for the sashimi plot
    :param fix_y_scale: bool, whether to fix y scale, default to True
    :param task_name: string, default is empty string, add this as a suffix to generated sashimi.pdf 

    :return cmd: string, the singularity cmd that you should type in the console when cd to the outdir

    Examples::

        root_dir = '/data/salomonis-archive/BAMs/NCI-R01/TCGA/TCGA-SKCM/TCGA_SKCM-BAMs/bams'
        bam_file_path = [root_dir+'/'+item+'.bam' for item in ['TCGA-3N-A9WB-06A-11R-A38C-07','TCGA-3N-A9WC-06A-11R-A38C-07','TCGA-3N-A9WD-06A-11R-A38C-07']]
        bai_file_path = [root_dir+'/'+item+'.bam.bai' for item in ['TCGA-3N-A9WB-06A-11R-A38C-07','TCGA-3N-A9WC-06A-11R-A38C-07','TCGA-3N-A9WD-06A-11R-A38C-07']]
        sif_anno_path = '/data/salomonis2/software/ggsashimi'
        cmd = snaf.downstream.prepare_sashimi_plot(bam_file_path,bai_file_path,'test',sif_anno_path,bam_contig_rename=[True,False,True],query_region='chr3:49099853-49099994')
        print(cmd)
        # singularity run -W $PWD -B $PWD:$PWD /data/salomonis2/software/ggsashimi/ggsashimi.sif -b input_bams.tsv -c chr3:49099853-49099994 -C 3 -M 3 --width 10 --fix-y-scale -g gencode.v36.annotation.gtf --ann-height=4 --height=2 -P palette.txt
    '''
    # copy file
    if not os.path.exists(outdir):
        os.mkdir(outdir)
    if not os.path.exists(os.path.join(outdir,'ggsashimi_bams')):
        os.mkdir(os.path.join(outdir,'ggsashimi_bams'))
    if not skip_copy:
        for bam in bam_path_list:
            print('copy {}'.format(bam))
            base = os.path.basename(bam)
            if not os.path.exists(os.path.join(outdir,'ggsashimi_bams',base)):
                subprocess.run(['cp','{}'.format(bam),'{}'.format(os.path.join(outdir,'ggsashimi_bams'))])
            else:
                print('{} exists, skip copy'.format(base))
        for bai in bai_path_list:
            print('copy {}'.format(bai))
            base = os.path.basename(bai)
            if not os.path.exists(os.path.join(outdir,'ggsashimi_bams',base)):
                subprocess.run(['cp','{}'.format(bai),'{}'.format(os.path.join(outdir,'ggsashimi_bams'))])
            else:
                print('{} exists, skip copy'.format(base))
        subprocess.run(['cp','{}'.format(os.path.join(sif_anno_path,'gencode.v36.annotation.gtf')),'{}'.format(outdir)])
        pwd = os.getcwd()
        os.chdir(os.path.join(outdir,'ggsashimi_bams'))
        all_bams = [os.path.basename(item) for item in bam_path_list]
        for boolean,bam in zip(bam_contig_rename,all_bams):
                if boolean:
                    cmd1 = 'samtools view -H {} | sed  -e \'s/SN:\([0-9XY]*\)/SN:chr\\1/\' -e \'s/SN:MT/SN:chrM/\' > in.header.sam'.format(bam)
                    print(cmd1)
                    subprocess.run(cmd1,shell=True)
                    cmd2 = 'samtools reheader in.header.sam {} > tmp.bam'.format(bam)
                    print(cmd2)
                    subprocess.run(cmd2,shell=True)
                    cmd3 = 'rm {}; rm {}; rm {}'.format(bam, 'in.header.sam','{}.bai'.format(bam))
                    print(cmd3)
                    subprocess.run(cmd3,shell=True)
                    cmd4 = 'mv tmp.bam {}'.format(bam)
                    print(cmd4)
                    subprocess.run(cmd4,shell=True)
                    cmd5 = 'samtools index {}'.format(bam)
                    print(cmd5)
                    subprocess.run(cmd5,shell=True)
        os.chdir(pwd)

    # create input_bams.tsv file
    with open(os.path.join(outdir,'input_bams.tsv'),'w') as f:
        for bam in bam_path_list:
            file_name = os.path.basename(bam)
            identifier = file_name.split('.bam')[0]
            path = os.path.join('ggsashimi_bams',file_name)
            category = identifier
            f.write('{}\t{}\t{}\n'.format(identifier,path,category))
    
    # create palette.txt file for standard igv color
    n = len(bam_path_list)
    colors = ['#FF3400','#3333FF','#00CD33','#B15117','#A445A8','#FF7700','#FF75C1','#999999']
    with open(os.path.join(outdir,'palette.txt'),'w') as f:
        for c in colors[:n]:
            f.write('{}\n'.format(c))
    if fix_y_scale:
        cmd = 'singularity run -W $PWD -B $PWD:$PWD {} -b input_bams.tsv -c {} -C 3 -M {} --width {} --fix-y-scale -g gencode.v36.annotation.gtf --ann-height={} --height={} -P palette.txt'.format(os.path.join(sif_anno_path,'ggsashimi.sif'),query_region,min_junction,width,ann_height,height)
    else:
        cmd = 'singularity run -W $PWD -B $PWD:$PWD {} -b input_bams.tsv -c {} -C 3 -M {} --width {} -g gencode.v36.annotation.gtf --ann-height={} --height={} -P palette.txt'.format(os.path.join(sif_anno_path,'ggsashimi.sif'),query_region,min_junction,width,ann_height,height)
    print(cmd)
    # execuate assuming singularity has been loaded
    old_pwd = os.getcwd()
    os.chdir(outdir)
    subprocess.run(cmd,shell=True)
    # rename the sashimi.pdf
    os.rename('sashimi.pdf','sashimi{}.pdf'.format(task_name))
    os.chdir(old_pwd)
    return cmd

    

    
def cross_survival_plot(survival1,survival2):
    pass
    






def mutation_analysis(mode,burden,mutation,output,n_sample_cutoff=10,gene_column='gene',genes_to_plot=None):
    '''
    Run mutation association analysis with neoantigen burden

    :param mode: string, either 'compute' or 'plot', compute will compute all the association, plot will visualize certain assciation in box plot
    :param burden: pandas dataframe, the burden result file generated by T pipeline, and sample ID should be consistent with the sample ID in mutation dataframe index
    :param mutation: pandas dataframe, the mutation data, make sure sample ID column is the index
    :param output: string, the path in which the result will be generated, if compute mode, a txt file will be generated, if plot mode, a plot will be generated
    :param n_sample_cutoff: int, default is 10, if a mutation occur in < n_sample_cutoff samples, it won't be considered
    :param gene_column: string, default is gene, in the mutation df, which column represents the gene
    :param gene_to_plot: list, each item is a gene name, for the plot mode

    Example::

        # compute mode
        snaf.mutation_analysis(mode='compute',burden=burden,mutation=mutation,output='result/stage3_mutation.txt')
        # plot mode
        snaf.mutation_analysis(mode='plot',burden=burden,mutation=mutation,output='result/stage3_mutation_CAMKK2.pdf',genes_to_plot=['CAMKK2'])
    '''
    if mode == 'compute':
        burden = burden.loc[:,burden.columns.isin(mutation.index)].iloc[-1]
        with open('snaf_mutation_tmp','w') as f:
            f.write('mutation_gene\tn_samples\tpval\n')
            for gene in tqdm(mutation['gene'].unique()):
                yes_samples = mutation.loc[mutation[gene_column] == gene,:].index.unique().tolist()
                if len(yes_samples) < n_sample_cutoff:
                    continue
                burden_df = burden.to_frame()
                burden_df.columns = ['burden']
                burden_df['mutation_{}'.format(gene)] = [True if sample in set(yes_samples) else False for sample in burden_df.index]
                x = burden_df.loc[burden_df['mutation_{}'.format(gene)],'burden'].values
                y = burden_df.loc[~(burden_df['mutation_{}'.format(gene)]),'burden'].values
                u,p = mannwhitneyu(x=x,y=y)
                f.write('{}\t{}\t{}\n'.format(gene,len(yes_samples),p))
        asso = pd.read_csv('snaf_mutation_tmp',sep='\t',index_col=0)
        results = ssm.multipletests(asso['pval'].values,alpha=0.05,method='fdr_bh')
        asso['adjp'] = results[1]
        asso.to_csv(output,sep='\t')
        os.remove('snaf_mutation_tmp')
    elif mode == 'plot':
        for gene in genes_to_plot:
            yes_samples = mutation.loc[mutation[gene_column] == gene,:].index.unique().tolist()
            burden_df = burden.loc[:,burden.columns.isin(mutation.index)].iloc[-1].to_frame()
            burden_df.columns = ['burden']
            burden_df['mutation_{}'.format(gene)] = [True if sample in set(yes_samples) else False for sample in burden_df.index]
            x = burden_df.loc[burden_df['mutation_{}'.format(gene)],'burden'].values
            y = burden_df.loc[~(burden_df['mutation_{}'.format(gene)]),'burden'].values
            u,p = mannwhitneyu(x=x,y=y)
            fig,ax = plt.subplots()
            sns.boxplot(data=burden_df,x='mutation_{}'.format(gene),y='burden',ax=ax,width=0.5)
            ax.text(x=0.5,y=0.5,s='mannwhitney p={}'.format(round(p,4)),weight='bold')
            plt.savefig('{}_{}.pdf'.format(output,gene),bbox_inches='tight')
            plt.close()



def survival_analysis(burden,survival,n,stratification_plot,survival_plot,
                      survival_duration='OS.time',survival_event='OS'): 
    '''
    Run survival analysis based on neoantigen burden, the `burden`and `survival`must be pandas dataframe, and make sure the 
    sample ID are consistent with each other, we only consider samples in `burden` that have clinical survival data.

    :param burden: pandas dataframe, the burden result file generated by T pipeline
    :param survival: pandas dataframe, the survival data
    :param n: int, how many groups to stratigy, support 2 (median), 3 (33%,66%), 4 (quantile)
    :param stratification_plot: string, the path and name for the stratification plot
    :param survival_plot: string, the path plus name for the survival plot
    :param survival_duration: string, which column in survival data represent the duration
    :param survival_event: string, which column in survival data represent the event

    Example::

        snaf.survival_analysis(burden,survival,n=2,stratification_plot='result/stage2_stratify.pdf',survival_plot='result/stage2_survival.pdf')

    '''

    burden = burden.loc[:,burden.columns.isin(survival.index)].iloc[-1]
    if n == 4:
        quantiles = burden.quantile([0.25,0.5,0.75]).values
        iqr = quantiles[2] - quantiles[0]
        upper_bound = quantiles[2] + 1.5*iqr
        lower_bound = quantiles[0] - 1.5*iqr
        identity_col = []
        for item in burden:
            if item > upper_bound:
                identity_col.append('outlier')
            elif item > quantiles[2] and item <= upper_bound:
                identity_col.append('high')
            elif item > quantiles[0] and item <= quantiles[2]:
                identity_col.append('medium')
            elif item >= lower_bound and item <= quantiles[0]:
                identity_col.append('low')
            elif item < lower_bound:
                identity_col.append('outlier')
    elif n == 3:
        quantiles = burden.quantile([0.33,0.66]).values
        identity_col = []
        for item in burden:
            if item > quantiles[1]:
                identity_col.append('high')
            elif item > quantiles[0] and item <= quantiles[1]:
                identity_col.append('medium')
            elif item <= quantiles[0]:
                identity_col.append('low')
    elif n == 2:
        quantiles = burden.quantile([0.25,0.5,0.75]).values
        iqr = quantiles[2] - quantiles[0]
        upper_bound = quantiles[2] + 1.5*iqr
        lower_bound = quantiles[0] - 1.5*iqr
        identity_col = []
        for item in burden:
            if item > upper_bound:
                identity_col.append('outlier')
            elif item > quantiles[1] and item <= upper_bound:
                identity_col.append('high')
            elif item >= lower_bound and item <= quantiles[1]:
                identity_col.append('low')
            else:
                identity_col.append('outlier')
    burden_output = burden.to_frame()
    burden_output['identity'] = identity_col
    burden_encode = pd.Series(index=burden.index,data=identity_col)
    be_vc = burden_encode.value_counts()
    # plot stratification
    sns.boxplot(data=burden.values);plt.savefig(stratification_plot,bbox_inches='tight');plt.close()
    high_group = burden_encode.loc[burden_encode=='high'].index.tolist()
    low_group = burden_encode.loc[burden_encode=='low'].index.tolist()
    high_os = survival.loc[high_group,[survival_duration,survival_event]]
    low_os = survival.loc[low_group,[survival_duration,survival_event]]
    # plot KM plot and logrank test
    fig,ax = plt.subplots()
    ax.set_ylim(0,1)
    for df in [low_os,high_os]:
        kmf = KaplanMeierFitter()
        kmf.fit(df[survival_duration],df[survival_event])
        kmf.plot_survival_function(ax=ax,ci_show=False,at_risk_counts=False)
    current_handles,current_labels = ax.get_legend_handles_labels()
    new_labels = ['low_burden','high_burden']
    ax.legend(current_handles,new_labels,bbox_to_anchor=(1,1),loc='upper left',frameon=False)
    results = logrank_test(low_os[survival_duration],high_os[survival_duration],low_os[survival_event],high_os[survival_event])
    ax.text(x=1000,y=0.05,s='Log-rank test: p-value is {:.2f}'.format(results.p_value),weight='bold')
    plt.savefig(survival_plot,bbox_inches='tight');plt.close()
    return burden_output,quantiles


def report_candidates(jcmq,df,sample,outdir,remove_quote=True,metrics={'netMHCpan_el':'binding_affinity','deepimmuno_immunogenicity':'immunogenicity'},
                      criterion=[('netMHCpan_el',0,'<=',2),('deepimmuno_immunogenicity',1,'==','True'),]):
    '''
    this function will report a txt file for all the T antigen candidate for a specific tumor sample.

    :param jcmq: the JunctionCountMatrixQuery object, usually you need to deserialize after_prediction.p file produced by T antigen pipeline
    :param df: DataFrame,the frequency table generated by T antigen pipeline
    :param sample: String, the sample name to report T antigen candiate
    :param outdir: string, the path to the output directory
    :param remove_quote: boolean, whether to remove the quotation or not, as one column in frequency table df is list, when loaded in memory using pandas, it will be added a quote, we can remove it
    :param metrics: dict, the key is the metric we registered to each peptide-hla, values is the corresponding name we want to output in the report for each metric
    :param criterion: list with each element as a tuple, this is a very specialized expression to specificy the criterion we are going to use to only report peptide-hla that meet the criterion.
                      each tuple has four element, the first element is the metric name we registered, the second is either 0 or 1, 0 means looking for the continuous value score,
                      1 means looking for the string identity value. See below note for further explanation. the third is an operator, and the fourth is the cutoff.

    .. note::

        By default, we run netMHCpan or MHCflurry for MHC binding prediction, and DeepImmuno for immunogenicity prediction. When we add the predicted value for
        each peptide-hla pair, we format the prediction result into a dataframe compliant with something like below:

          peptide     mer     hla       score    identity
         AAAAAAAAA     9   HLA-A*01:01   0.3       SB  

        here is the dataframe for MHC binding prediction, we have a column recording continous value (score), and another column (identity) recording whether it is
        a SB (Strong Binder) or WB (Weaker Binder). Those are all useful information coming out of the predictor. Now we just need to register this df to the NeoJunction
        class by running ``nj.enhanced_peptide.register_attr(df,'netMHCpan_el)``. We can keep adding scores to the NeoJunction class. This explain what the 0/1 in criterion
        mean.

    Example::

        snaf.report_candidates(jcmq,df,'SRR067783.bed','result',True)
        # a txt file will be written to the outdir you specified

    '''
    if not os.path.exists(outdir):
        os.mkdir(outdir)
    df = df.copy()
    from ast import literal_eval
    if remove_quote:
        df['samples'] = [literal_eval(item) for item in df['samples']]
    # build dict for retrieving specificity scores and occurence
    df_score = df.filter(like='tumor_specificity',axis=1)
    score_dict = {}   # {tumor_specificity_mle:{aa1_uid1:0.5,aa2_uid2:0.6},tumor_specifity_bayesian:{}...}
    for col in df_score.columns.tolist() + ['n_sample','coord','symbol']:
        tmp = df[col].to_dict()
        score_dict[col] = tmp
    # start to process jcmq
    junction_count_dict = jcmq.junction_count_matrix[sample].to_dict()
    col_index = jcmq.subset.columns.tolist().index(sample)
    results = jcmq.results[0]
    hlas = jcmq.results[1]
    selected_hla = hlas[col_index]
    with open(os.path.join(outdir,'T_antigen_candidates_{}.txt'.format(sample)),'w') as f:
        f.write('sample\tpeptide\tuid\tjunction_count\tphase\tevidences\thla\t')
        metrics_stream = '\t'.join(list(metrics.values())) + '\t'
        f.write(metrics_stream)
        score_stream = '\t'.join(list(score_dict.keys())) + '\n'
        f.write(score_stream)
        for item,samples in zip(df.index,df['samples']):
            if sample in samples:
                stream = '{}\t'.format(sample)
                aa,uid = item.split(',')
                jc = junction_count_dict[uid]
                stream += '{}\t{}\t{}\t'.format(aa,uid,jc)
                row_index = jcmq.subset.index.tolist().index(uid)
                nj = deepcopy(results[row_index])
                nj.enhanced_peptides = nj.enhanced_peptides.filter_based_on_hla(selected_hla=selected_hla)
                ep = nj.enhanced_peptides.filter_based_on_criterion(criterion,True)  # only report valid hla
                origin = ep[len(aa)][aa]['origin']
                stream += '{}\t{}\t'.format(origin[2],origin[3])  # phase, evidences
                for hla in ep[len(aa)][aa].keys():
                    if hla != 'origin':
                        stream += '{}\t'.format(hla)
                        for k,v in metrics.items():
                            s = nj.enhanced_peptides[len(aa)][aa][hla][k][0]
                            stream += '{}\t'.format(s)
                        for k,v in score_dict.items():
                            s = v[item]
                            stream += '{}\t'.format(s)
                        f.write(stream.rstrip('\t') + '\n')
                        stream = '\t'.join(stream.split('\t')[:6]) + '\t'



    


def stage0_compatible_results(jcmq,outdir='.',name_burden='burden_stage0.txt',name_frequency='frequency_stage0.txt'):
    # for burden
    burden_stage0 = jcmq.cond_df.astype('int8')
    burden_stage0.loc['burden'] = burden_stage0.sum(axis=0).values
    burden_stage0['mean'] = burden_stage0.mean(axis=1).values
    burden_stage0.to_csv(os.path.join(outdir,name_burden),sep='\t')
    # for frequency
    tuple_list = []
    for index,series in jcmq.cond_df.iterrows():
        samples = series.loc[series].index.tolist()
        tuple_ = (index,samples,len(samples))
        tuple_list.append(tuple_)
    df = pd.DataFrame.from_records(tuple_list,columns=['junction','samples','n_sample']).set_index(keys='junction').sort_values(by='n_sample',ascending=False)
    df.to_csv(os.path.join(outdir,name_frequency),sep='\t')


def add_gene_symbol_frequency_table(df,remove_quote=True):
    '''
    This function will convert the ENSG id to gene sysmbol and add a column in place for your dataframe

    :param df: the input df, make sure the index is the uid
    :param remove_quote: bool, depending on how your df is loaded, if directly read from the disk, certain column will contain quotation, whether to remove it or not

    :return df: the output df, with added gene symbol column

    Example::

        add_gene_symbol_frequency_table(df=df,remove_quote=False)
    '''
    # the index has to be comma separated string
    from ast import literal_eval
    if remove_quote:
        df['samples'] = [literal_eval(item) for item in df['samples']]
    ensg_list = [item.split(',')[1].split(':')[0] for item in df.index] 
    symbol_list = ensemblgene_to_symbol(ensg_list,'human') 
    df['symbol'] = symbol_list
    return df


def reformat_frequency_table(df,remove_quote=True):
    # the index has to be comma separated string
    from ast import literal_eval
    if remove_quote:
        df['samples'] = [literal_eval(item) for item in df['samples']]
    sequence_io = []
    for row in df.itertuples():
        for item in row.samples:
            sequence_io.append((row.Index,item,1))
    result = pd.DataFrame.from_records(sequence_io,columns=['id','sample','value'])
    result = result.groupby(by=['id','sample'])['value'].mean().unstack(fill_value=0)
    return result
        

def ensemblgene_to_symbol(query,species):
    '''
    Examples::
        from sctriangulate.preprocessing import GeneConvert
        converted_list = GeneConvert.ensemblgene_to_symbol(['ENSG00000010404','ENSG00000010505'],species='human')
    '''
    # assume query is a list, will also return a list
    import mygene
    mg = mygene.MyGeneInfo()
    out = mg.querymany(query,scopes='ensemblgene',fileds='symbol',species=species,returnall=True,as_dataframe=True,df_index=True)

    df = out['out']
    df_unique = df.loc[~df.index.duplicated(),:]
    df_unique['symbol'].fillna('unknown_gene',inplace=True)
    mapping = df_unique['symbol'].to_dict()

    result = []
    for item in query:
        result.append(mapping[item])

    return result


def analyze_neoantigens(freq_path,junction_path,total_samples,outdir,fasta=False,mers=None,columns=None,cutoffs=(0.1,0.9),junction_bl=0.1):
    '''
    Users need to execute this function to prepare input for the t antigen dash viewer

    :param freq_path: string, the path to the frequency file in the result folder, with uid appended
    :param junction_path: string, the path to the burden0 file in the result folder
    :param total_samples: int, the total number of samples, will be used for determine shared versus unique neoantigen
    :param outdir: string, where the output will go into
    :param fasta: bool, whether to output the fasta file as well to check the motif 
    :param mers: None or list, if [9,10], it will generate mer9 and mer10 for fasta and separate dataframe.
    :param columns: None or string, which column to include in the output df
    :param cutoffs: tuple, the cutoffs to determine shared versus unique neoantigen
    :param junction_bl: float, the cutff for parent junction, if junction occur < junction_bl sample, we don't consider neoantigens from those junction

    Example::

        # just output for viewer
        snaf.downstream.analyze_neoantigens(freq_path='result/frequency_stage2_verbosity1_uid.txt',junction_path='result/burden_stage0.txt',total_samples=472,outdir='result',mers=None,fasta=False)
        # to run MEME for motif after that
        snaf.downstream.analyze_neoantigens(freq_path='result/frequency_stage2_verbosity1_uid.txt',junction_path='result/burden_stage0.txt',total_samples=472,outdir='result',mers=[9,10],fasta=True)
    '''
    freq = pd.read_csv(freq_path,sep='\t',index_col=0)
    junction = pd.read_csv(junction_path,sep='\t',index_col=0)['mean'].to_dict()
    freq['uid'] = [item.split(',')[-1] for item in freq.index]
    freq.index = [item.split(',')[0] for item in freq.index]
    freq['mean_percent_samples_junction_present'] = freq['uid'].map(junction).values
    freq['actual_percent_samples_neoantigen_present'] = freq['n_sample'] / total_samples
    freq.drop(columns='samples',inplace=True)
    identity = []
    for e,o in zip(freq['mean_percent_samples_junction_present'],freq['actual_percent_samples_neoantigen_present']):
        if o < cutoffs[0] * e:
            identity.append('low')
        elif o > cutoffs[1] * e:
            identity.append('high')
        else:
            identity.append('medium')
    freq['identity'] = identity
    freq = freq.loc[np.logical_not(freq.index.duplicated()),:]
    freq = freq.loc[freq['identity']!='medium',:]
    freq = freq.loc[freq['mean_percent_samples_junction_present']>junction_bl,:]
    freq['length'] = [len(item) for item in freq.index]
    if columns is None:
        columns = []
    selected_columns = ['n_sample','uid','mean_percent_samples_junction_present','actual_percent_samples_neoantigen_present','identity','length'] + columns
    freq = freq.loc[:,selected_columns]
    sns.regplot(data=freq,x='mean_percent_samples_junction_present',y='actual_percent_samples_neoantigen_present',fit_reg=False,x_jitter=0.01,y_jitter=0.01,scatter_kws={'s':5})
    plt.savefig(os.path.join(outdir,'shared_vs_unique_neoantigen_all.pdf'),bbox_inches='tight')
    plt.close()
    if mers is None:
        freq.to_csv(os.path.join(outdir,'shared_vs_unique_neoantigen_all.txt'),sep='\t')
    else:
        for mer in mers:
            freq_mer = freq.loc[freq['length']==mer,:]
            freq_mer.to_csv(os.path.join(outdir,'shared_vs_unique_neoantigen_mer{}.txt'.format(mer)),sep='\t')
            if fasta:
                with open(os.path.join(outdir,'mer{}_high.fasta'.format(mer)),'w') as f1, open(os.path.join(outdir,'mer{}_low.fasta'.format(mer)),'w') as f2:  
                    for identity,sub_df in freq_mer.groupby(by='identity'):
                        if identity == 'high':
                            for item in sub_df.index:
                                f1.write('>{}\n{}\n'.format(item,item))
                        else:
                            for item in sub_df.index:
                                f2.write('>{}\n{}\n'.format(item,item))  
                     









    


import matplotlib as mpl
mpl.use('Agg')
from matplotlib.patches import PathPatch
import matplotlib.pyplot as plt
from scipy.stats import entropy
from collections import Counter
import seaborn as sns
import numpy as np
import argparse
from .schema import *
import os,sys

import matplotlib as mpl
mpl.rcParams['pdf.fonttype'] = 42
mpl.rcParams['ps.fonttype'] = 42
mpl.rcParams['font.family'] = 'Arial'


def draw_letter(letter,x,y,x_scale=1,y_scale=1,ax=None):
    text = letter_schema[letter]
    transform = mpl.transforms.Affine2D().scale(x_scale*globscale,y_scale*globscale) + \
                mpl.transforms.Affine2D().translate(x,y) + \
                ax.transData
    p = PathPatch(text,lw=0,fc=color_schema[letter],transform=transform)
    ax.add_artist(p)

def draw_weblogo(all_score):
    fig,ax = plt.subplots()
    x = 1
    highest = 0
    for score in all_score:
        y = 0
        for b,s in score:
            draw_letter(b,x,y,x_scale=1,y_scale=s,ax=ax)
            #raise Exception
            y += s
        x += 1.5
        highest = max(highest,y)
    ax.set_xlim((0,x))
    ax.set_ylim((0,highest))
    ax.set_ylabel('bits')
    ax.set_xticks(np.linspace(1,1+(len(all_score)-1)*1.5,len(all_score)))
    ax.set_xticklabels(['pos{}'.format(i) for i in range(1,len(all_score)+1)])
    ax.set_xlabel('position')
    sns.despine(ax=ax,offset=10,trim=True)

def calculate_weblogo(pep_list):
    amino = 'ARNDCQEGHILKMFPSTWYV'
    n = len(pep_list)
    l = len(pep_list[0])
    all_score = []
    all_bit = []
    for i,pos in enumerate(zip(*pep_list)):
        score = []
        counter = Counter(pos)
        freq = np.array(list(counter.values()))/n
        ent = np.log2(20) - entropy(freq,base=2)
        for aa in amino:
            score.append((aa,ent*counter.get(aa,0)/n))
        score = sorted(score,key=lambda x:x[1])
        all_score.append(score)
        all_bit.append(ent)
    return all_score,all_bit

def run_pweblogo(pep_list,out):
    all_score,all_bit = calculate_weblogo(pep_list)
    draw_weblogo(all_score)
    plt.savefig(out,bbox_inches='tight')
    plt.close()





import requests
import sys,os
import subprocess
import re

def run_ensembl(ens):
    # return the string for sequence, either full_length (using ENST) or peptide (using ENSP)
    server = "https://rest.ensembl.org"
    ext = "/sequence/id/{}?".format(ens)
    r = requests.get(server+ext, headers={ "Content-Type" : "text/plain"})
    return r.text



def run_emboss(asequence,bsequence,python_executable):
    # return the string for alignment
    current_pwd = os.getcwd()
    os.chdir(os.path.dirname(__file__))
    command = '{} emboss.py --email whatever@gmail.com --stype protein --asequence {} --bsequence {}'.format(python_executable,asequence,bsequence)
    capture_pat = re.compile(r'Creating result file: (.*?.out.txt)')
    stdout = subprocess.run(command,shell=True,stdout=subprocess.PIPE,universal_newlines=True).stdout
    outfile = re.search(capture_pat,stdout).group(1)
    with open('{}'.format(outfile),'r') as f:
        string_stream = ''
        f_list = f.readlines()
        for i,line in enumerate(f_list):
            if line.startswith('EMBOSS_001'):
                string_stream += f_list[i]
                string_stream += f_list[i+1]
    string_stream = '\n'.join(string_stream.split('\n')[:-1])
    clear_commands = ['rm *.txt','rm *.params']
    for clear_command in clear_commands:
        subprocess.run(clear_command,shell=True)
    os.chdir(current_pwd)
    return string_stream
    








#!/data/salomonis2/LabFiles/Frank-Li/refactor/neo_env/bin/python3.7

import numpy as np
import pandas as pd
import os
import sys
import pickle
import h5py
import matplotlib.pyplot as plt
import anndata as ad
from scipy.optimize import minimize, minimize_scalar
from scipy import stats
from scipy.sparse import csr_matrix, find
from tqdm import tqdm
import re

try:
    import pymc3 as pm   # conda install -c conda-forge pymc3 mkl-service
    import theano
    import arviz as az
except ImportError:
    print('''
        Optional package pymc3 is not installed, it is for calculating tumor specificity using hirerarchical bayesian model
        For Linux: https://github.com/pymc-devs/pymc/wiki/Installation-Guide-(Linux)
        For MacOS: https://github.com/pymc-devs/pymc/wiki/Installation-Guide-(MacOS)
        For PC:    https://github.com/pymc-devs/pymc/wiki/Installation-Guide-(Windows)
    ''')

'''
this script is to query the tumor specificity of the junction
'''


def gtex_configuration(df,gtex_db,t_min_arg,n_max_arg,normal_cutoff_arg,tumor_cutoff_arg,normal_prevalance_cutoff_arg,tumor_prevalance_cutoff_arg,add_control=None):
    global adata_gtex
    global adata
    global t_min
    global n_max
    global normal_cutoff
    global tumor_cutoff
    global normal_prevalance_cutoff
    global tumor_prevalance_cutoff
    tested_junctions = set(df.index)
    adata = ad.read_h5ad(gtex_db)
    adata = adata[np.logical_not(adata.obs_names.duplicated()),:] 
    adata = adata[list(set(adata.obs_names).intersection(tested_junctions)),:]  
    print('Current loaded gtex cohort with shape {}'.format(adata.shape))
    tissue_dict = adata.var['tissue'].to_dict()
    adata_gtex = adata   # already has mean and tissue variables
    if add_control is not None:
        for id_, control in add_control.items():
            if isinstance(control,pd.DataFrame):
                assert len(set(control.columns).intersection(tissue_dict.keys())) == 0  # sample id can not be ambiguous
                control = control.loc[np.logical_not(control.index.duplicated()),:]
                control = control.loc[list(set(control.index).intersection(tested_junctions)),:]
                print('Adding cohort {} with shape {} to the database'.format(id_,control.shape))
                tissue_dict_right = {k:id_ for k in control.columns}
                tissue_dict.update(tissue_dict_right)
                df_left = adata.to_df()
                df_right = control
                df_combine = df_left.join(other=df_right,how='outer').fillna(0)
                adata = ad.AnnData(X=csr_matrix(df_combine.values),obs=pd.DataFrame(index=df_combine.index),var=pd.DataFrame(index=df_combine.columns))

            elif isinstance(control,ad.AnnData):
                assert len(set(control.var_names).intersection(tissue_dict.keys())) == 0
                control = control[np.logical_not(control.obs_names.duplicated()),:]
                control = control[list(set(control.obs_names).intersection(tested_junctions)),:]
                print('Adding cohort {} with shape {} to the database'.format(id_,control.shape))
                if 'tissue' in control.var.columns:   # if tissue is in var columns, it will be used 
                    tissue_dict_right = control.var['tissue'].to_dict()
                else:
                    tissue_dict_right = {k:id_ for k in control.var_names}
                tissue_dict.update(tissue_dict_right)
                df_left = adata.to_df()
                df_right = control.to_df()
                df_combine = df_left.join(other=df_right,how='outer').fillna(0)
                adata = ad.AnnData(X=csr_matrix(df_combine.values),obs=pd.DataFrame(index=df_combine.index),var=pd.DataFrame(index=df_combine.columns))
            
            else:
                raise Exception('control must be either in dataframe or anndata format')

            print('now the shape of control db is {}'.format(adata.shape))
            adata.var['tissue'] = adata.var_names.map(tissue_dict).values
            adata.obs['mean'] = np.array(adata.X.mean(axis=1)).squeeze()
            total_count = np.array(adata.X.sum(axis=0)).squeeze() / 1e6
            adata.var['total_count'] = total_count
            

    t_min = t_min_arg
    n_max = n_max_arg
    normal_cutoff = normal_cutoff_arg
    tumor_cutoff = tumor_cutoff_arg
    normal_prevalance_cutoff = normal_prevalance_cutoff_arg
    tumor_prevalance_cutoff = tumor_prevalance_cutoff_arg

    return adata


def multiple_crude_sifting(junction_count_matrix,add_control,dict_exonlist,outdir,filter_mode):
    if filter_mode == 'prevalance':
        valid,invalid,cond_df = multiple_crude_sifting_prevalance(junction_count_matrix,add_control,dict_exonlist,outdir)
    elif filter_mode == 'maxmin':
        valid,invalid,cond_df = multiple_crude_sifting_maxmin(junction_count_matrix,add_control,dict_exonlist,outdir)
    return valid,invalid, cond_df

def multiple_crude_sifting_prevalance(junction_count_matrix,add_control=None,dict_exonlist=None,outdir='.'):

    if not os.path.exists(outdir):
        os.mkdir(outdir)

    df_to_write = []
    df = pd.DataFrame(index=junction_count_matrix.index)
    prevalance_tumor = np.count_nonzero((junction_count_matrix > tumor_cutoff).values,axis=1) / junction_count_matrix.shape[1]
    df['prevalance_tumor'] = prevalance_tumor
    # consider gtex
    prevalance_normal = np.count_nonzero((adata_gtex.X > normal_cutoff).toarray(),axis=1) / adata_gtex.shape[1]
    prevalance_normal_dict = {j:v for j,v in zip(adata_gtex.obs_names,prevalance_normal)}
    df['prevalance_normal'] = df.index.map(prevalance_normal_dict).fillna(value=0)
    df['cond'] = (df['prevalance_tumor'] > tumor_prevalance_cutoff) & (df['prevalance_normal'] < normal_prevalance_cutoff)
    valid = df.loc[df['cond']].index.tolist()
    tmp = df.copy()
    df_to_write.append(tmp)
    print('reduce valid NeoJunction from {} to {} because they are present in GTEx'.format(df.shape[0],len(valid)))
    if dict_exonlist is not None:   # a valid junction can not be present in any ensembl documented transcript
        updated_valid = []
        for uid in tqdm(valid):
            ensg = uid.split(':')[0]
            exons = ':'.join(uid.split(':')[1:])
            if '_' in exons or 'U' in exons or 'ENSG' in exons or 'I' in exons:
                updated_valid.append(uid)
            else:
                exonlist = dict_exonlist[ensg]
                exonstring = '|'.join(exonlist)
                e1,e2 = exons.split('-')
                pattern1 = re.compile(r'^{}\|{}\|'.format(e1,e2))  # ^E1.1|E2.3|
                pattern2 = re.compile(r'\|{}\|{}$'.format(e1,e2))  # |E1.1|E2.3$
                pattern3 = re.compile(r'\|{}\|{}\|'.format(e1,e2)) # |E1.1|E2.3|
                if re.search(pattern3,exonstring) or re.search(pattern2,exonstring) or re.search(pattern1,exonstring):   # as long as match one pattern, should be eliminated
                    continue
                else:
                    updated_valid.append(uid)
        print('reduce valid Neojunction from {} to {} because they are present in Ensembl db'.format(len(valid),len(updated_valid)))
        valid = updated_valid
    # consider add_control
    if add_control is not None:
        for i,(id_,control) in enumerate(add_control.items()):
            n_previous_valid = len(valid)
            if isinstance(control,pd.DataFrame):
                prevalance_normal = np.count_nonzero((control > normal_cutoff).values,axis=1) / control.shape[1]
                prevalance_normal_dict = {j:v for j,v in zip(control.index, prevalance_normal)}
            elif isinstance(control,ad.AnnData):
                prevalance_normal = np.count_nonzero((control.X > normal_cutoff).toarray(),axis=1) / control.shape[1]
                prevalance_normal_dict = {j:v for j,v in zip(control.obs_names,prevalance_normal)}
            else:
                raise Exception('control must be either in dataframe or anndata format')
            df['prevalance_normal_add'] = df.index.map(prevalance_normal_dict).fillna(value=0)
            df['cond_add'] = (df['prevalance_tumor'] > tumor_prevalance_cutoff) & (df['prevalance_normal_add'] < normal_prevalance_cutoff)
            valid_add = df.loc[df['cond_add']].index.tolist()
            valid = list(set(valid).intersection(set(valid_add)))
            tmp = df.copy(); tmp.drop(columns=['prevalance_tumor','prevalance_normal','cond'],inplace=True); tmp.rename(columns=lambda x:x+'_{}'.format(id_),inplace=True)
            df_to_write.append(tmp)
            print('reduce valid Neojunction from {} to {} because they are present in added control {}'.format(n_previous_valid,len(valid),id_))
    invalid = list(set(junction_count_matrix.index).difference(set(valid)))
    # now, consider each entry
    t_min = tumor_cutoff
    valid_set = set(valid)
    cond_dict = {j:(True if j in valid_set else False) for j in junction_count_matrix.index}
    tmp = pd.DataFrame(index=junction_count_matrix.index,data={'placeholder':junction_count_matrix.index.map(cond_dict).values})
    first_half_cond_df = pd.concat([tmp]*junction_count_matrix.shape[1],axis=1)
    first_half_cond_df.columns = junction_count_matrix.columns
    cond_df = (first_half_cond_df) & (junction_count_matrix > t_min)
    # write the df
    df_to_write = pd.concat(df_to_write,axis=1)
    df_to_write.to_csv(os.path.join(outdir,'NeoJunction_statistics_prevalance.txt'),sep='\t')
    return valid,invalid,cond_df



def multiple_crude_sifting_maxmin(junction_count_matrix,add_control=None,dict_exonlist=None,outdir='.'):   # for JunctionCountMatrixQuery class, only consider gtex
    if not os.path.exists(outdir):
        os.mkdir(outdir)
        
    df = pd.DataFrame(index=junction_count_matrix.index,data = {'max':junction_count_matrix.max(axis=1).values})
    df_to_write = []
    # consider gtex
    junction_to_mean = adata_gtex.obs.loc[adata_gtex.obs_names.isin(junction_count_matrix.index),'mean'].to_dict()
    df['mean'] = df.index.map(junction_to_mean).fillna(value=0)
    df['diff'] = df['max'] - df['mean']
    df['cond'] = (df['mean'] < n_max) & (df['diff'] > t_min)
    valid = df.loc[df['cond']].index.tolist()
    tmp = df.copy()
    df_to_write.append(tmp)
    print('reduce valid NeoJunction from {} to {} because they are present in GTEx'.format(df.shape[0],len(valid)))
    if dict_exonlist is not None:   # a valid junction can not be present in any ensembl documented transcript
        updated_valid = []
        for uid in tqdm(valid):
            ensg = uid.split(':')[0]
            exons = ':'.join(uid.split(':')[1:])
            if '_' in exons or 'U' in exons or 'ENSG' in exons or 'I' in exons:
                updated_valid.append(uid)
            else:
                exonlist = dict_exonlist[ensg]
                exonstring = '|'.join(exonlist)
                e1,e2 = exons.split('-')
                pattern1 = re.compile(r'^{}\|{}\|'.format(e1,e2))  # ^E1.1|E2.3|
                pattern2 = re.compile(r'\|{}\|{}$'.format(e1,e2))  # |E1.1|E2.3$
                pattern3 = re.compile(r'\|{}\|{}\|'.format(e1,e2)) # |E1.1|E2.3|
                if re.search(pattern3,exonstring) or re.search(pattern2,exonstring) or re.search(pattern1,exonstring):   # as long as match one pattern, should be eliminated
                    continue
                else:
                    updated_valid.append(uid)
        print('reduce valid Neojunction from {} to {} because they are present in Ensembl db'.format(len(valid),len(updated_valid)))
        valid = updated_valid
    # consider add_control
    mean_add_list = []
    if add_control is not None:
        for i,(id_,control) in enumerate(add_control.items()):
            n_previous_valid = len(valid)
            if isinstance(control,pd.DataFrame):
                junction_to_mean = control.mean(axis=1).to_dict()
            elif isinstance(control,ad.AnnData):
                junction_to_mean = control.to_df().mean(axis=1).to_dict()
            else:
                raise Exception('control must be either in dataframe or anndata format')
            df['mean_add'] = df.index.map(junction_to_mean).fillna(value=0)
            df['diff_add'] = df['max'] - df['mean_add']
            df['cond_add'] = (df['mean_add'] < n_max) & (df['diff_add'] > t_min)
            mean_add_list.append(df['mean_add'])
            valid_add = df.loc[df['cond_add']].index.tolist()
            valid = list(set(valid).intersection(set(valid_add)))
            tmp = df.copy(); tmp.drop(columns=['mean','diff','cond'],inplace=True); tmp.rename(columns=lambda x:x+'_{}'.format(id_),inplace=True)
            df_to_write.append(tmp)
            print('reduce valid Neojunction from {} to {} because they are present in added control {}'.format(n_previous_valid,len(valid),id_))
    invalid = list(set(junction_count_matrix.index).difference(set(valid)))
    # now, consider each entry
    gtex_df = pd.concat([df['mean']]*junction_count_matrix.shape[1],axis=1)
    gtex_df.columns = junction_count_matrix.columns
    diff_df_gtex = junction_count_matrix - gtex_df
    cond_df = (gtex_df < n_max) & (diff_df_gtex > t_min)
    if add_control is not None:
        for mean_add in mean_add_list:
            add_df = pd.concat([mean_add]*junction_count_matrix.shape[1],axis=1)    
            add_df.columns = junction_count_matrix.columns
            diff_df_add = junction_count_matrix - add_df
            cond_df = cond_df & (add_df < n_max) & (diff_df_add > t_min)
    # write the df
    df_to_write = pd.concat(df_to_write,axis=1)
    df_to_write.to_csv(os.path.join(outdir,'NeoJunction_statistics_maxmin.txt'),sep='\t')
    return valid,invalid,cond_df


def crude_tumor_specificity(uid,count):    # for NeoJunction class, since we normally start from Jcmq with check_gtex=False, rarely being called.
    detail = ''
    if uid not in set(adata.obs_names):
        mean_value = 0
    else:
        mean_value = adata.obs.loc[uid,'mean']
    diff = count - mean_value
    if mean_value < n_max and diff >= t_min:
        identity = True
    else:
        identity = False
    return identity,mean_value


def mle_func(parameters,y):
    sigma = parameters
    ll = np.sum(stats.halfnorm.logpdf(y,0,sigma))
    neg_ll = -1 * ll
    return neg_ll

def split_df_to_chunks(df,cores=None):
    df_index = np.arange(df.shape[0])
    if cores is None:
        cores = mp.cpu_count()
    sub_indices = np.array_split(df_index,cores)
    sub_dfs = [df.iloc[sub_index,:] for sub_index in sub_indices]
    return sub_dfs


def split_array_to_chunks(array,cores=None):
    if not isinstance(array,list):
        raise Exception('split_array_to_chunks function works for list, not ndarray')
    array_index = np.arange(len(array))
    if cores is None:
        cores = mp.cpu_count()
    sub_indices = np.array_split(array_index,cores)
    sub_arrays = []
    for sub_index in sub_indices:
        item_in_group = []
        for i in sub_index:
            item_in_group.append(array[i])
        sub_arrays.append(item_in_group)
    return sub_arrays


def add_tumor_specificity_frequency_table(df,method='mean',remove_quote=True,cores=None):
    '''
    add tumor specificty to each neoantigen-uid in the frequency table produced by SNAF T pipeline

    :param df: DataFrame, the frequency table produced by SNAF T pipeline
    :param method: string, either 'mean', or 'mle', or 'bayesian'
    :param remove quote: boolean, whether to remove the quotation or not, as one column in frequency table df is list, when loaded in memory using pandas, it will be added a quote, we can remove it
    :param cores: int, how many cpu cores to use for this computation, default None and use all the cpu the program detected

    :return new_df: a dataframe with one added column containing tumor specificity score

    Example::

        snaf.add_tumor_specificity_frequency_table(df,'mle',remove_quote=True)

    '''
    from ast import literal_eval
    import multiprocessing as mp
    if remove_quote:
        df['samples'] = [literal_eval(item) for item in df['samples']]
    if cores is None:
        cores = mp.cpu_count()

    if method != 'bayesian': 
        pool = mp.Pool(processes=cores)
        print('{} subprocesses have been spawned'.format(cores))

        all_unique_junctions = list(set([item.split(',')[1] for item in df.index]))
        sub_arrays = split_array_to_chunks(all_unique_junctions,cores=cores)
        r = [pool.apply_async(func=add_tumor_specificity_frequency_table_atomic_func,args=(sub_array,method,)) for sub_array in sub_arrays]  

        pool.close()
        pool.join()
        results = []
        for collect in r:
            result = collect.get()
            results.append(result)
        all_score_dict = {}
        for score_dict in results:
            all_score_dict.update(score_dict)
        col = []
        for item in df.index:
            col.append(all_score_dict[item.split(',')[1]])
        new_df = df.copy()
        new_df['tumor_specificity_{}'.format(method)] = col

    else:   # seems like bayesian doesn't work well with multiprocessing
        all_unique_junctions = list(set([item.split(',')[1] for item in df.index]))
        score_dict = {}
        for uid in tqdm(all_unique_junctions,total=len(all_unique_junctions)):
            score_dict[uid] = tumor_specificity(uid,'bayesian')
        for item in df.index:
            col.append(score_dict[item.split(',')[1]])
        new_df = df.copy()
        new_df['tumor_specificity_{}'.format(method)] = col
            

    return new_df

def add_tumor_specificity_frequency_table_atomic_func(sub_array,method):
    uid_list = sub_array
    score_dict = {uid:tumor_specificity(uid,method) for uid in tqdm(uid_list,total=len(uid_list))}
    return score_dict


    

def tumor_specificity(uid,method,return_df=False):
    try:
        info = adata[[uid],:]
    except:
        print('{} not detected in gtex, impute as zero'.format(uid))
        info = ad.AnnData(X=csr_matrix(np.full((1,adata.shape[1]),0)),obs=pd.DataFrame(data={'mean':[0]},index=[uid]),var=adata.var)  # weired , anndata 0.7.6 can not modify the X in place? anndata 0.7.2 can do that in scTriangulate
    df = pd.DataFrame(data={'value':info.X.toarray().squeeze(),'tissue':info.var['tissue'].values},index=info.var_names)
    if method == 'mean':
        try:
            sigma = adata.obs.loc[uid,'mean']
        except KeyError:
            sigma = 0
        if return_df:
            return sigma,df
        else:
            return sigma
    elif method == 'mle':
        scale_factor_dict = adata.var['total_count'].to_dict()
        df['value_cpm'] = df['value'].values / df.index.map(scale_factor_dict).values
        y = df['value_cpm'].values
        # mle_model = minimize(mle_func,np.array([0.2]),args=(y,),bounds=((0,1),),method='Nelder-Mead')
        mle_model = minimize_scalar(mle_func,bounds=(0,1),args=(y,),method='bounded')
        if mle_model.success:
            sigma = mle_model.x
        else:   
            sigma = 0
            print(uid,y, mle_model)   # debug purpose
        if return_df:
            return sigma,df
        else:
            return sigma
    elif method == 'bayesian':
        scale_factor_dict = adata.var['total_count'].to_dict()
        df['value_cpm'] = df['value'].values / df.index.map(scale_factor_dict).values
        y = df['value_cpm'].values
        x = []
        for tissue in adata.var['tissue'].unique():
            sub = adata[uid,adata.var['tissue']==tissue]
            total_count = sub.shape[1]
            c = np.count_nonzero(sub.X.toarray())
            scaled_c = round(c * (25/total_count),0)
            x.append(scaled_c)
        x = np.array(x)
        try:
            with pm.Model() as m:
                sigma = pm.Uniform('sigma',lower=0,upper=1)
                nc = pm.HalfNormal('nc',sigma=sigma,observed=y)
                nc_hat = pm.Deterministic('nc_hat',pm.math.sum(nc)/len(y))
                psi = pm.Beta('psi',alpha=2,beta=nc_hat*20)
                mu = pm.Gamma('mu',alpha=nc_hat*50,beta=1)
                c = pm.ZeroInflatedPoisson('c',psi,mu,observed=x)
                trace = pm.sample(draws=1000,step=pm.NUTS(target_accept=0.95),tune=1000,return_inferencedata=False,cores=1)
                '''
                the error of "Got error No model on context stack. trying to find log_likelihood in translation" maybe due to pymc build and how they launch multi-cores.
                remember, my build can only work when cores=1, which further indicate there might be an issue revolving around it.
                https://stackoverflow.com/questions/69888492/sampling-of-pymc3-in-python-gets-runtime-error-of-bootstrapping-phase
                '''

            df = az.summary(trace,round_to=2)

            '''
            az.summary(trace)

                    mean    sd  hdi_3%  hdi_97%  mcse_mean  mcse_sd  ess_bulk  ess_tail  r_hat
            sigma    0.47  0.01    0.46     0.48       0.00     0.00    182.23     98.84   1.02
            nc_hat   0.22  0.00    0.22     0.22       0.00     0.00    200.00    200.00    NaN
            mu      22.97  0.52   21.87    23.87       0.04     0.03    196.69     94.84   1.00

            az.plot_posterior(trace,var_names=['sigma','nc_hat','mu'])
            az.plot_forest(trace,,var_names=['sigma','nc_hat','mu'])

            gv = pm.model_to_graphviz(m)
            gv.format = 'pdf'
            gv.render(filename='model_graph');sys.exit('stop')
            # to run the above, you need to module load graphviz so that dot is exposed to the program

            '''
            sigma = df.iloc[0]['mean']

        except:
            sigma = None
            print(uid,x,y)

        if return_df:
            return sigma,df
        else:
            return sigma
            



















#!/data/salomonis2/LabFiles/Frank-Li/refactor/neo_env/bin/python3.7

import numpy as np
import pandas as pd
import os
import sys
import pickle
import h5py
import matplotlib.pyplot as plt
import xmltodict
import copy
from tqdm import tqdm

from Bio.SeqIO.FastaIO import SimpleFastaParser

########################## Following is for manipulating fasta db files
def chop_normal_pep_db(fasta_path,output_path,mers,allow_duplicates):
    '''
    chop any normal human proteome to certain mers

    :param fasta_path: string, the path to the human protein fasta file
    :param output_path: string, the path to the output fasta file
    :param mers: list, like [9,10] will generate 9mer and 10mer
    :param allow_duplicates: boolean. whether allow duplicate or not

    Example::

        chop_normal_pep_db(fasta_path='human_uniprot_proteome.fasta',output_path='./human_uniprot_proteome_mer9_10.fasta',mers=[9,10],allow_duplicates=False)
    '''
    # for human genome in uniprot, 9-10mer, remove duplicates will decrease from 44,741,578 to 41,638,172
    if allow_duplicates:
        with open(fasta_path,'r') as in_handle, open(output_path,'w') as out_handle:
            for title,seq in tqdm(SimpleFastaParser(in_handle)):
                count = 0
                length = len(seq)
                for i in range(length):
                    for mer in mers:
                        if i+mer <= length:
                            out_handle.write('>{}_{}_{}\n'.format(title,mer,count))
                            out_handle.write('{}\n'.format(seq[i:i+mer]))
                            count += 1
    else:
        with open(fasta_path,'r') as in_handle, open(output_path,'w') as out_handle:
            existing = set()
            for title,seq in tqdm(SimpleFastaParser(in_handle)):
                count = 0
                length = len(seq)
                for i in range(length):
                    for mer in mers:
                        if i+mer <= length:
                            subseq = seq[i:i+mer]
                            if subseq not in existing:                                
                                out_handle.write('>{}_{}_{}\n'.format(title,mer,count))
                                out_handle.write('{}\n'.format(subseq))
                                existing.add(subseq)
                                count += 1        

def compare_two_fasta(fa1_path,fa2_path,outdir='.',write_comm=False,write_unique1=False,write_unique2=False,prefix=''):
    '''
    Compare arbitracy two fasta files, report unique and common ones

    :param fa1_path: the first fasta file path that you want to compare with
    :param fa2_path: the second fasta file path that you want to compare with
    :param write_comm: boolean, whether write the common one between fa1 and fa2
    :param write_unique1: boolean, whether write the one unique to fa1
    :param write_unique2: boolean, whether write the one unique to fa2
    :param prefix: string, whether to add a prefix to the output fasta file name

    Example::

        snaf.proteomics.compare_two_fasta(fa1_path='./fasta/human_proteome_uniprot_9_10_mers_unique.fasta', 
                                      fa2_path='./fasta/neoantigen_{}_unique.fasta'.format(sample),outdir='./fasta',
                                      write_unique2=True,prefix='{}_'.format(sample))

    '''
    seq1,seq2 = set(),set()
    seq1_l,seq2_l = [],[]
    seq1_t,seq2_t = [],[]
    for i,path in enumerate([fa1_path,fa2_path]):
        with open(path,'r') as f:
            for t,s in SimpleFastaParser(f):
                if i == 0:
                    if s not in seq1:
                        seq1.add(s)
                        seq1_l.append(s)
                        seq1_t.append(t)
                elif i == 1:
                    if s not in seq2:
                        seq2.add(s)
                        seq2_l.append(s)
                        seq2_t.append(t)
    comm = seq1.intersection(seq2)
    unique1 = seq1.difference(seq2)
    unique2 = seq2.difference(seq1)
    print('common:{}\nunique1:{}\nunique2:{}'.format(len(comm),len(unique1),len(unique2)))
    if write_comm:
        print('writing comm')
        with open(os.path.join(outdir,'{}comm.fasta'.format(prefix)),'w') as f:
            for item in comm:
                seq1_list = seq1_l
                item_t = seq1_t[seq1_list.index(item)]
                f.write('>{}\n{}\n'.format(item_t,item))
    if write_unique1:
        print('writing unique1')
        with open(os.path.join(outdir,'{}unique1.fasta'.format(prefix)),'w') as f:
            for item in unique1:
                seq1_list = seq1_l
                item_t = seq1_t[seq1_list.index(item)]
                f.write('>{}\n{}\n'.format(item_t,item))
    if write_unique2:
        print('writing unique2')
        with open(os.path.join(outdir,'{}unique2.fasta'.format(prefix)),'w') as f:
            for item in unique2:
                seq2_list = seq2_l
                item_t = seq2_t[seq2_list.index(item)]
                f.write('>{}\n{}\n'.format(item_t,item))




def remove_redundant(fasta_path,out_path):
    '''
    remove redundant entries in any fasta files

    :param fasta_path: string, the path to the input fasta file
    :param out_path: string, the path to the output folder

    Example::

        snaf.proteomics.remove_redundant('./fasta/neoantigen_{}.fasta'.format(sample),'./fasta/neoantigen_{}_unique.fasta'.format(sample))
    '''
    existing = set()
    original_count = 0
    with open(fasta_path,'r') as f1, open(out_path,'w') as f2:
        for t,s in SimpleFastaParser(f1):
            original_count += 1
            if s not in existing:
                f2.write('>{}\n{}\n'.format(t,s))
                existing.add(s)
    print('reduce from {} down to {}'.format(original_count,len(existing)))
                

######################## Marks the end of fasta db manipulation

            
#######################  following functions are all for setting mqpar.xml file for maxQuant
def add_database_file(doc,dbs):
    for db in dbs:
        try:
            template = copy.deepcopy(doc['MaxQuantParams']['fastaFiles']['FastaFileInfo'][-1]) # non-first time
        except:
            template = copy.deepcopy(doc['MaxQuantParams']['fastaFiles']['FastaFileInfo']) # first time
        finally:
            template['fastaFilePath'] = db
            template['identifierParseRule'] = '>(.*)'
            template['descriptionParseRule'] = '>(.*)'
            try:
                doc['MaxQuantParams']['fastaFiles']['FastaFileInfo'].append(template)
            except:
                doc['MaxQuantParams']['fastaFiles']['FastaFileInfo'] = []
                doc['MaxQuantParams']['fastaFiles']['FastaFileInfo'].append(template)
    return doc

def numThreads(doc,thread):
    doc['MaxQuantParams']['numThreads'] = str(thread)
    return doc

def add_input_files(doc,inputs):
    experiment = 1
    for inp in inputs:
        # 'filePaths'
        if not type(doc['MaxQuantParams']['filePaths']['string']) == list:
            doc['MaxQuantParams']['filePaths']['string'] = []
            doc['MaxQuantParams']['filePaths']['string'].append(inp)
        else:
            doc['MaxQuantParams']['filePaths']['string'].append(inp)
        
        # 'experiments'
        if not type(doc['MaxQuantParams']['experiments']['string']) == list:
            doc['MaxQuantParams']['experiments']['string'] = []
            doc['MaxQuantParams']['experiments']['string'].append(experiment)
        else:
            doc['MaxQuantParams']['experiments']['string'].append(experiment)
        experiment += 1
        
        # 'fractions'
        if not type(doc['MaxQuantParams']['fractions']['short']) == list:
            doc['MaxQuantParams']['fractions']['short'] = []
            doc['MaxQuantParams']['fractions']['short'].append('32767')
        else:
            doc['MaxQuantParams']['fractions']['short'].append('32767')
            
        # 'ptms'
        if not type(doc['MaxQuantParams']['ptms']['boolean']) == list:
            doc['MaxQuantParams']['ptms']['boolean'] = []
            doc['MaxQuantParams']['ptms']['boolean'].append('False')
        else:
            doc['MaxQuantParams']['ptms']['boolean'].append('False')
            
        # 'paramGroupIndices'
        if not type(doc['MaxQuantParams']['paramGroupIndices']['int']) == list:
            doc['MaxQuantParams']['paramGroupIndices']['int'] = []
            doc['MaxQuantParams']['paramGroupIndices']['int'].append('0')
        else:
            doc['MaxQuantParams']['paramGroupIndices']['int'].append('0')
            
        # 'referenceChannel'
        if not type(doc['MaxQuantParams']['referenceChannel']['string']) == list:
            doc['MaxQuantParams']['referenceChannel']['string'] = []
            doc['MaxQuantParams']['referenceChannel']['string'].append(None)
        else:
            doc['MaxQuantParams']['referenceChannel']['string'].append(None)
    
    return doc

def change_enzymes(doc,enzymes,mode):

    '''
    mode 0: specific
    mode 3: semi-specific
    mode 4: unspecific
    mode 5: no digestion
    '''

    doc['MaxQuantParams']['parameterGroups']['parameterGroup']['enzymeMode'] = str(mode)

    if enzymes == None:
        doc['MaxQuantParams']['parameterGroups']['parameterGroup']['enzymes'] = None
    else:
        for enzyme in enzymes:
            if not type(doc['MaxQuantParams']['parameterGroups']['parameterGroup']['enzymes']['string']) == list:
                doc['MaxQuantParams']['parameterGroups']['parameterGroup']['enzymes']['string'] = []
                doc['MaxQuantParams']['parameterGroups']['parameterGroup']['enzymes']['string'].append(enzyme)
            else:
                doc['MaxQuantParams']['parameterGroups']['parameterGroup']['enzymes']['string'].append(enzyme)
     
    return doc           


def change_fdr(doc,protein_fdr,peptide_fdr,site_fdr):
    doc['MaxQuantParams']['proteinFdr'] = protein_fdr
    doc['MaxQuantParams']['peptideFdr'] = peptide_fdr
    doc['MaxQuantParams']['siteFdr'] = site_fdr
    return doc

def change_contaminants(doc,include):
    doc['MaxQuantParams']['includeContaminants'] = str(include)
    return doc

def change_variable_modifications(doc,mods=['Oxidation (M)', 'Acetyl (Protein N-term)', 'Carbamidomethyl (C)']):
    if mods is not None:
        doc['MaxQuantParams']['parameterGroups']['parameterGroup']['variableModifications']['string'] = mods
    else:
        doc['MaxQuantParams']['parameterGroups']['parameterGroup']['variableModifications'] = mods
    return doc

def change_fixed_modifications(doc,mods=None):
    if mods is not None:
        doc['MaxQuantParams']['parameterGroups']['parameterGroup']['fixedModifications']['string'] = []
        for mod in mods:
            doc['MaxQuantParams']['parameterGroups']['parameterGroup']['fixedModifications']['string'].append(mod)
    return doc

def change_length(doc,minPepLen,maxPeptideMass,minPeptideLengthForUnspecificSearch,maxPeptideLengthForUnspecificSearch):
    doc['MaxQuantParams']['minPepLen'] = minPepLen
    doc['MaxQuantParams']['maxPeptideMass'] = maxPeptideMass
    doc['MaxQuantParams']['minPeptideLengthForUnspecificSearch'] = minPeptideLengthForUnspecificSearch
    doc['MaxQuantParams']['maxPeptideLengthForUnspecificSearch'] = maxPeptideLengthForUnspecificSearch
    return doc


def set_maxquant_configuration(dbs,n_threads,inputs,enzymes,enzyme_mode,outdir,
                               outname='mqpar.xml',protein_fdr=0.01,peptide_fdr=0.01,site_fdr=0.01,include_contaminants=True,var_mods=['Oxidation (M)', 'Acetyl (Protein N-term)', 'Carbamidomethyl (C)'],fix_mods=None,
                               minPepLen=8,maxPeptideMass=4600,minPeptideLengthForUnspecificSearch=8,maxPeptideLengthForUnspecificSearch=25):
    '''
    automatically build MaxQuant configuration file mqpar.xml, the template mqpar.xml is generated using maxquant 2.0.3.1 on a windows PC GUI,
    we load a single raw file, a single fasta file, reduce carbamidomethyl from fixed modification to variable modification and allow match run, otherwise,
    we retain all default parameters.

    :param dbs: list, each element is the path to the database fasta file
    :param n_threads: int, how many threads will be used for MaxQuant search
    :param inputs: list, each element is the path to the raw file
    :param enzymes: list, what enzymes will be used for search
    :param enzyme_mode: int, 0 -> specific, 3 -> semi-specific, 4 -> unspecific, 5 -> no digestion
    :param outdir: string, the output directory for the conf file
    :param outname: string, default is mqpar.xml
    :param protein_fdr: float, default is 0.01
    :param peptide_fdr: float, default is 0.01
    :param site_fdr: float, default is 0.01
    :param include_contaminants: boolean, default is True
    :param var_mods: list or None, default is ['Oxidation (M)', 'Acetyl (Protein N-term)', 'Carbamidomethyl (C)']
    :param fix_mods: list or None, default is []
    :param minPepLen: int, default is 8
    :param maxPeptideMass: int, default is 4600
    :param minPeptideLengthForUnspecificSearch: int, default is 8
    :param maxPeptideLengthForUnspecificSearch: int, default is 25

    Example::

            dbs = ['/data/salomonis2/LabFiles/Frank-Li/neoantigen/MS/schuster/RNA/snaf_analysis/fasta/SRR5933726.Aligned.sortedByCoord.out.bed_unique2.fasta']
            inputs = ['/data/salomonis2/LabFiles/Frank-Li/neoantigen/MS/schuster/MS/OvCa48/OvCa48_classI_Rep#1.raw',
                    '/data/salomonis2/LabFiles/Frank-Li/neoantigen/MS/schuster/MS/OvCa48/OvCa48_classI_Rep#2.raw',
                    '/data/salomonis2/LabFiles/Frank-Li/neoantigen/MS/schuster/MS/OvCa48/OvCa48_classI_Rep#3.raw',
                    '/data/salomonis2/LabFiles/Frank-Li/neoantigen/MS/schuster/MS/OvCa48/OvCa48_classI_Rep#4.raw',
                    '/data/salomonis2/LabFiles/Frank-Li/neoantigen/MS/schuster/MS/OvCa48/OvCa48_classI_Rep#5.raw',
                    '/data/salomonis2/LabFiles/Frank-Li/neoantigen/MS/schuster/MS/OvCa48/OvCa48_classI_Rep#6.raw']
            outdir = '/data/salomonis2/LabFiles/Frank-Li/neoantigen/MS/schuster/MS/OvCa48'
            snaf.proteomics.set_maxquant_configuration(dbs=dbs,n_threads=20,inputs=inputs,enzymes=None,enzyme_mode=5,protein_fdr=1,peptide_fdr=0.05,site_fdr=1,outdir=outdir)

    '''
    with open (os.path.join(os.path.dirname(__file__),'mqpar.xml'),'r') as fr:
        doc = xmltodict.parse(fr.read()) 
        doc = add_database_file(doc,dbs)
        doc = numThreads(doc,n_threads)
        doc = add_input_files(doc,inputs)        
        doc = change_enzymes(doc,enzymes,enzyme_mode)
        doc = change_fdr(doc,protein_fdr=protein_fdr,peptide_fdr=peptide_fdr,site_fdr=site_fdr)
        doc = change_contaminants(doc,include_contaminants)
        doc = change_variable_modifications(doc,var_mods)
        doc = change_fixed_modifications(doc,fix_mods)
        doc = change_length(doc,minPepLen,maxPeptideMass,minPeptideLengthForUnspecificSearch,maxPeptideLengthForUnspecificSearch)
        a = xmltodict.unparse(doc,pretty=True,encoding='utf-8')
        a = a.replace('&gt;','>')
        with open(os.path.join(outdir,outname),'w') as fw:
            fw.write(a)

        '''

        Download the maxQuant (working version, 1.6.14 and 2.0.3.1) zip from official website, on windowns, just lanuch exe, follow the instructions,
        .Net framework should be fine, if .Net core 3.1 is needed, just install it and verify it in cmd. On Linux, Using below instructions.

        SAMPLE=OvCa114
        cd /data/salomonis2/LabFiles/Frank-Li/neoantigen/revision/ovarian/MS/${SAMPLE}
        module load mono/5.20.1
        export PATH=$PATH:​/usr/local/mono/5.20.1/bin
        mono /data/salomonis2/LabFiles/Frank-Li/neoantigen/MS/MaxQuant_2.0.3.1/bin/MaxQuantCmd.exe /data/salomonis2/LabFiles/Frank-Li/neoantigen/revision/ovarian/MS/${SAMPLE}/mqpar.xml 

        '''
#######################  Here marks the end of maxQuant configuration


def summarize_ms_result(peptide,msms,freq,outdir):
    '''
    After finishing running Maxquant, this function can help figuring out what peptide for further validation

    :param peptide: string, the path to MaxQuant peptide.txt file
    :param msms: string, the path to the MaxQuant msms.txt file
    :param freq: string, the path to the SNAF frequency_stage2_verbosity1_uid_gene_symbol_coord_mean_mle.txt file
    :param outdir: string, the path to the output folder

    :return df_peptide: dataframe, the valid MS-evidenced neoantigens with associated properties

    Examples::

        df = snaf.proteomics.summarize_ms_result(peptide='MS/OvCa114/combined/txt/peptides.txt',msms='MS/OvCa114/combined/txt/msms.txt',freq='result/frequency_stage2_verbosity1_uid_gene_symbol_coord_mean_mle.txt')

    '''
    df_peptide = pd.read_csv(peptide,sep='\t',index_col=0)
    df_peptide = df_peptide.loc[df_peptide['Proteins'].notna(),:].sort_values(by='PEP')
    # add msms
    df_msms = pd.read_csv(msms,sep='\t',index_col='id')
    dic_msms_matches = df_msms['Matches'].to_dict()
    dic_msms_intensities = df_msms['Intensities'].to_dict()
    df_peptide['best_msms_matches'] = df_peptide['Best MS/MS'].map(dic_msms_matches)
    df_peptide['best_msms_intensities'] = df_peptide['Best MS/MS'].map(dic_msms_intensities)
    # add freq
    df_freq = pd.read_csv(freq,sep='\t',index_col=0)
    df_freq['pep'] = [item.split(',')[0] for item in df_freq.index]
    df_freq['uid'] = [item.split(',')[1] for item in df_freq.index]
    from ast import literal_eval
    df_freq['samples'] = [literal_eval(item) for item in df_freq['samples']]
    dic_pep_uid = {}
    dic_pepuid_property = {}
    for row in df_freq.itertuples():
        dic_pep_uid.setdefault(row.pep,[]).append(row.uid)
        dic_pepuid_property[row.Index] = (row.samples,row.n_sample,row.symbol,row.coord,row.tumor_specificity_mean,row.tumor_specificity_mle)
    df_peptide['uids'] = df_peptide.index.map(dic_pep_uid).values
    df_peptide['properties'] = [[dic_pepuid_property[','.join([pep,uid])] for uid in row] for row,pep in zip(df_peptide['uids'],df_peptide.index)]
    df_peptide.to_csv(os.path.join(outdir,'summaried_ms_results.txt'),sep='\t')
    return df_peptide




import os,sys
from setuptools import setup

# build long description
base_dir = os.path.dirname(os.path.abspath(__file__))
long_description = '\n\n'.join([open(os.path.join(base_dir,'README.md'),'r').read()])

# requires
requires = [
    'tensorflow==2.3.0',
    'pandas==1.3.4',
    'numpy==1.18.5',
    'numba==0.53.0',
    'mhcflurry==2.0.5',
    'h5py==2.10.0',
    'anndata==0.7.6',
    'seaborn==0.11.2',
    'biopython==1.79',
    'requests==2.26.0',
    'xmltodict==0.12.0',
    'xmltramp2==3.1.1',
    'tqdm==4.62.3',
    'scipy==1.4.1',
    'statsmodels==0.13.1',
    'lifelines==0.26.4',
    'umap-learn==0.5.2',
    'plotly==5.4.0',
    'Werkzeug==2.0.2',
    'dash==2.0.0',
    'dash-dangerously-set-inner-html==0.0.2',
    'mygene==3.2.2'
]

setup(
    name = 'SNAF',
    version = '0.7.0',
    description = 'A Python package to predict, prioritize and visualize splicing derived neoantigens, including MHC-bound peptides (T cell antigen) and altered surface protein (B cell antigen)',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    author = 'Guangyuan(Frank) Li',
    author_email='li2g2@mail.uc.edu',
    maintainer='Guangyuan(Frank) Li',
    maintainer_email = 'li2g2@mail.uc.edu',
    url='https://github.com/frankligy/SNAF',
    project_urls={
        'Documentation':'https://snaf.readthedocs.io',
    },
    packages=['snaf','snaf/deepimmuno','snaf/dash_app','snaf/surface'],
    package_data = {'snaf/deepimmuno':['data/*','models/cnn_model_331_3_7/.data-00000-of-00001','models/cnn_model_331_3_7/.index','models/cnn_model_331_3_7/checkpoint'],
                    'snaf':['mqpar.xml','HLA_Allele_frequency_21_populations.csv']},
    install_requires=requires,
    python_requires='>=3.7',
    classifers=[
          'Development Status :: 3 - Alpha',
          'Programming Language :: Python :: 3',
          'License :: OSI Approved :: BSD License',
          'Operating System :: OS Independent',
    ]
)




'''
Program to run deepimmuno-cnn
'''
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
import warnings
warnings.filterwarnings('ignore')
import logging
logger = logging.getLogger()   # this is a singleton, all .tf .keras logger will pass the info to this module level rootlogger
logger.setLevel(level=logging.ERROR)
import tensorflow as tf
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
import tensorflow.keras as keras
from tensorflow.keras import layers
import numpy as np
import pandas as pd
import argparse





def seperateCNN():
    input1 = keras.Input(shape=(10, 12, 1))
    input2 = keras.Input(shape=(46, 12, 1))

    x = layers.Conv2D(filters=16, kernel_size=(2, 12))(input1)  # 9
    x = layers.BatchNormalization()(x)
    x = keras.activations.relu(x)
    x = layers.Conv2D(filters=32, kernel_size=(2, 1))(x)    # 8
    x = layers.BatchNormalization()(x)
    x = keras.activations.relu(x)
    x = layers.MaxPool2D(pool_size=(2, 1), strides=(2, 1))(x)  # 4
    x = layers.Flatten()(x)
    x = keras.Model(inputs=input1, outputs=x)

    y = layers.Conv2D(filters=16, kernel_size=(15, 12))(input2)     # 32
    y = layers.BatchNormalization()(y)
    y = keras.activations.relu(y)
    y = layers.MaxPool2D(pool_size=(2, 1), strides=(2, 1))(y)  # 16
    y = layers.Conv2D(filters=32,kernel_size=(9,1))(y)    # 8
    y = layers.BatchNormalization()(y)
    y = keras.activations.relu(y)
    y = layers.MaxPool2D(pool_size=(2, 1),strides=(2,1))(y)  # 4
    y = layers.Flatten()(y)
    y = keras.Model(inputs=input2,outputs=y)

    combined = layers.concatenate([x.output,y.output])
    z = layers.Dense(128,activation='relu')(combined)
    z = layers.Dropout(0.2)(z)
    z = layers.Dense(1,activation='sigmoid')(z)

    model = keras.Model(inputs=[input1,input2],outputs=z)
    return model

def pull_peptide_aaindex(dataset):
    result = np.empty([len(dataset),10,12,1])
    for i in range(len(dataset)):
        result[i,:,:,:] = dataset[i][0]
    return result


def pull_hla_aaindex(dataset):
    result = np.empty([len(dataset),46,12,1])
    for i in range(len(dataset)):
        result[i,:,:,:] = dataset[i][1]
    return result


def pull_label_aaindex(dataset):
    col = [item[2] for item in dataset]
    result = [0 if item == 'Negative' else 1 for item in col]
    result = np.expand_dims(np.array(result),axis=1)
    return result

def pull_label_aaindex(dataset):
    result = np.empty([len(dataset),1])
    for i in range(len(dataset)):
        result[i,:] = dataset[i][2]
    return result

def aaindex(peptide,after_pca):

    amino = 'ARNDCQEGHILKMFPSTWYV-'
    matrix = np.transpose(after_pca)   # [12,21]
    encoded = np.empty([len(peptide), 12])  # (seq_len,12)
    for i in range(len(peptide)):
        query = peptide[i]
        if query == 'X': query = '-'
        query = query.upper()
        encoded[i, :] = matrix[:, amino.index(query)]

    return encoded


def peptide_data_aaindex(peptide,after_pca):   # return numpy array [10,12,1]
    length = len(peptide)
    if length == 10:
        encode = aaindex(peptide,after_pca)
    elif length == 9:
        peptide = peptide[:5] + '-' + peptide[5:]
        encode = aaindex(peptide,after_pca)
    encode = encode.reshape(encode.shape[0], encode.shape[1], -1)
    return encode

def dict_inventory(inventory):
    dicA, dicB, dicC = {}, {}, {}
    dic = {'A': dicA, 'B': dicB, 'C': dicC}

    for hla in inventory:
        type_ = hla[4]  # A,B,C
        first2 = hla[6:8]  # 01
        last2 = hla[8:]  # 01
        try:
            dic[type_][first2].append(last2)
        except KeyError:
            dic[type_][first2] = []
            dic[type_][first2].append(last2)

    return dic


def rescue_unknown_hla(hla, dic_inventory):
    type_ = hla[4]
    first2 = hla[6:8]
    last2 = hla[8:]
    big_category = dic_inventory[type_]
    #print(hla)
    if not big_category.get(first2) == None:
        small_category = big_category.get(first2)
        distance = [abs(int(last2) - int(i)) for i in small_category]
        optimal = min(zip(small_category, distance), key=lambda x: x[1])[0]
        return 'HLA-' + str(type_) + '*' + str(first2) + str(optimal)
    else:
        small_category = list(big_category.keys())
        distance = [abs(int(first2) - int(i)) for i in small_category]
        optimal = min(zip(small_category, distance), key=lambda x: x[1])[0]
        return 'HLA-' + str(type_) + '*' + str(optimal) + str(big_category[optimal][0])






def hla_data_aaindex(hla_dic,hla_type,after_pca,dic_inventory):    # return numpy array [46,12,1]
    try:
        seq = hla_dic[hla_type]
    except KeyError:
        hla_type = rescue_unknown_hla(hla_type,dic_inventory)
        seq = hla_dic[hla_type]
    encode = aaindex(seq,after_pca)
    encode = encode.reshape(encode.shape[0], encode.shape[1], -1)
    return encode

def construct_aaindex(ori,hla_dic,after_pca,dic_inventory):
    series = []
    for i in range(ori.shape[0]):
        peptide = ori['peptide'].iloc[i]
        hla_type = ori['HLA'].iloc[i]
        immuno = np.array(ori['immunogenicity'].iloc[i]).reshape(1,-1)   # [1,1]

        encode_pep = peptide_data_aaindex(peptide,after_pca)    # [10,12]

        encode_hla = hla_data_aaindex(hla_dic,hla_type,after_pca,dic_inventory)   # [46,12]
        series.append((encode_pep, encode_hla, immuno))
    return series

def hla_df_to_dic(hla):
    dic = {}
    for i in range(hla.shape[0]):
        col1 = hla['HLA'].iloc[i]  # HLA allele
        col2 = hla['pseudo'].iloc[i]  # pseudo sequence
        dic[col1] = col2
    return dic



def computing_s(peptide,mhc):

    after_pca = np.loadtxt('./data/after_pca.txt')
    hla = pd.read_csv('./data/hla2paratopeTable_aligned.txt',sep='\t')
    hla_dic = hla_df_to_dic(hla)
    inventory = list(hla_dic.keys())
    dic_inventory = dict_inventory(inventory)
    cnn_model = seperateCNN()
    cnn_model.load_weights('./models/cnn_model_331_3_7/')
    peptide_score = [peptide]
    hla_score = [mhc]
    immuno_score = ['0']
    ori_score = pd.DataFrame({'peptide':peptide_score,'HLA':hla_score,'immunogenicity':immuno_score})
    dataset_score = construct_aaindex(ori_score,hla_dic,after_pca,dic_inventory)
    input1_score = pull_peptide_aaindex(dataset_score)
    input2_score = pull_hla_aaindex(dataset_score)
    label_score = pull_label_aaindex(dataset_score)
    scoring = cnn_model.predict(x=[input1_score,input2_score])
    return float(scoring)


def file_process(df_input):

    after_pca = np.loadtxt(os.path.join(os.path.dirname(os.path.abspath(__file__)),'data/after_pca.txt'))
    hla = pd.read_csv(os.path.join(os.path.dirname(os.path.abspath(__file__)),'data/hla2paratopeTable_aligned.txt'),sep='\t')
    hla_dic = hla_df_to_dic(hla)
    inventory = list(hla_dic.keys())
    dic_inventory = dict_inventory(inventory)
    cnn_model = seperateCNN()
    cnn_model.load_weights(os.path.join(os.path.dirname(os.path.abspath(__file__)),'models/cnn_model_331_3_7/'))

    ori_score = df_input
    ori_score.columns = ['peptide', 'HLA']
    ori_score['immunogenicity'] = ['0'] * ori_score.shape[0]
    dataset_score = construct_aaindex(ori_score, hla_dic, after_pca, dic_inventory)

    input1_score = pull_peptide_aaindex(dataset_score)
    input2_score = pull_hla_aaindex(dataset_score)
    label_score = pull_label_aaindex(dataset_score)
    scoring = cnn_model.predict(x=[input1_score, input2_score])
    scoring = cnn_model.predict(x=[input1_score, input2_score])
    ori_score['immunogenicity'] = scoring
    
    return ori_score







import pandas as pd
import numpy as np
import os,sys
import subprocess
import multiprocessing
import re

def alignment_to_uniprot(orf,uid,dict_uni_fa,tmhmm=False,software_path=None):
    ensgid = uid.split(':')[0]
    isoforms = dict_uni_fa[ensgid]  # {acc1:seq,acc1-2:seq}
    results = []
    for o in orf:
        if o != 'unrecoverable' and o != '':
            final_notes = []
            for iso,iso_seq in isoforms.items():
                if len(iso_seq) != len(o):
                    conclusion = True
                else:
                    notes = []                    
                    for i, mer in enumerate(chop_sequence(o,10)):
                        if mer in iso_seq:
                            notes.append(True)  # here True means align
                        else:
                            notes.append(False)
                    conclusion = not all(notes)   # here True means not existing before
                final_notes.append(conclusion)
            decision = all(final_notes)  # here True means this isoform is novel
            if tmhmm:
                is_cross_membrane = TMHMM(o,software_path=software_path)
                decision = all([decision,is_cross_membrane])
            results.append(decision)
        else:
            results.append(o)
    return results
    

def chop_sequence(seq,kmer):   # how to splice sequence, elegant way to use range
    frag_bucket = []
    for i in range(0,len(seq),kmer):
        if i + kmer <= len(seq):
            frag_bucket.append(seq[i:i+kmer])
        else:
            frag_bucket.append(seq[i:])
    return frag_bucket


def TMHMM(aa,software_path):
    # download TMHMM linux version, untar it.
    # change shabang of tmhmm and tmhmmformat.pl to the path of perl 5+ you loaded
    # export the path of tmhmm to $PATH, finsh configuration

    # in my use case, save those for convenience
    # perl: /usr/local/perl/5.20.1/bin/perl   # no need to load perl when running tmhmm
    # tmhmm: /data/salomonis2/LabFiles/Frank-Li/python3/TMHMM/tmhmm-2.0c/bin
    name = os.getpid()
    int_file_path = os.path.join(os.path.dirname(os.path.dirname(__file__)),'scratch','{}.fasta'.format(name))
    with open(int_file_path,'w') as f1:
        f1.write('>peptide_{}\n'.format(name))
        f1.write(aa)
    '''
    # sp|Q9NR97|TLR8_HUMAN Length: 942
    # sp|Q9NR97|TLR8_HUMAN Number of predicted TMHs:  1
    # sp|Q9NR97|TLR8_HUMAN Exp number of AAs in TMHs: 23.20959
    # sp|Q9NR97|TLR8_HUMAN Exp number, first 60 AAs:  0.03874
    # sp|Q9NR97|TLR8_HUMAN Total prob of N-in:        0.00259
    sp|Q9NR97|TLR8_HUMAN	TMHMM2.0	outside	     1   825
    sp|Q9NR97|TLR8_HUMAN	TMHMM2.0	TMhelix	   826   848
    sp|Q9NR97|TLR8_HUMAN	TMHMM2.0	inside	   849   942
    '''
    lines = subprocess.run('{} {}'.format(software_path,int_file_path),shell=True,stdout=subprocess.PIPE,universal_newlines=True).stdout.split('\n')[:-1]
    n_cross = int(lines[1].split('  ')[-1])
    result = True if n_cross > 0 else False
    return result